import logging
from typing import Dict

class ColorFormatter(logging.Formatter):
    """Custom formatter that adds colors to log levels

    Attributes:
        COLOR_MAP (Dict[str, str]): A dictionary mapping log levels to their corresponding colors.

    Methods:
        Public Methods:
            format: Format the log record with colors.
    """

    COLOR_MAP: Dict[str, str] = {
        "DEBUG": "\033[36m",  # Cyan
        "INFO": "\033[32m",  # Green
        "WARNING": "\033[33m",  # Yellow
        "ERROR": "\033[31m",  # Red
        "CRITICAL": "\033[41m",  # Red background
        "RESET": "\033[0m",  # Reset color
    }

    def format(self, record: logging.LogRecord) -> str:
        """Format the log record with colors

        Args:
            record (logging.LogRecord): The log record to format.

        Returns:
            str: The formatted log record with colors.
        """
        # Color only for console output (StreamHandler)
        handler = logging.getLogger(record.name).handlers[0]
        if type(handler) is logging.StreamHandler:
            levelname = record.levelname
            if levelname in self.COLOR_MAP:
                record.levelname = f"{self.COLOR_MAP[levelname]}{levelname}{self.COLOR_MAP['RESET']}"
        return super().format(record)


# Keep track of configured loggers to
# avoid re-configuring the same logger multiple times
# for seperate class instances.
#
# Implements a singleton pattern for loggers of each unique name.
#
# Results in only one logger instance per unique name (i.e. a class object)
# per runtime instance of the program.
_configured_loggers: Dict[str, logging.Logger] = {}

# Log levels for export across the package
# so that other modules can use them without
# needing to import the logging module.
DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

# Used to validate the log_level argument is
# a valid python logging log level
# in the get_logger() function.
VALID_LOG_LEVELS = {DEBUG, INFO, WARNING, ERROR, CRITICAL}

def _validate_log_level(level: int) -> None:
    """Validate the log level is a valid python logging log level

    Args:
        level (logging.Level): The log level to validate.

    Returns:
        None
    """
    if level not in VALID_LOG_LEVELS:
        raise ValueError(
            f"Invalid log level provided: `{level}`. "
            f"Must be one of: {VALID_LOG_LEVELS}."
        )

def _is_existing_logger(module_name: str) -> bool:
    """Check if a logger already exists for the module

    Args:
        module_name (str): The name of the module to check for an existing logger.

    Returns:
        bool: True if the logger exists, False otherwise.
    """
    if module_name in _configured_loggers:
        return True
    return False

def _get_existing_logger(module_name: str) -> logging.Logger:
    """Get the existing logger for the module

    Args:
        module_name (str): The name of the module to get the logger for.

    Returns:
        logging.Logger: The existing logger for the module.
    """
    return _configured_loggers[module_name]

def _get_real_logger(
    module_name: str,
    level: int | None = WARNING,
) -> logging.Logger:
    """Configure a logger for a specific module.

    Creates a new, or returns an existing, logger for the module.
    If there is an existing logger for the module, and it is a null logger,
    a real logger will be created and returned.
    
    Args:
        module_name (str): The name of the module to configure logging for.
            This should be passed in as the `__name__` variable of the module.
        level (logging.Level | None): The log level to use for the module.
            Defaults to logging.WARNING.

    Returns:
        logging.Logger: The configured logger for the module.
    """
    if level is None:
        level = WARNING
    
    if _is_existing_logger(module_name):
        existing_logger = _get_existing_logger(module_name)
        # Check if any of the handlers is a NullHandler
        if not any(
            isinstance(handler, logging.NullHandler) 
            for handler in existing_logger.handlers
        ):
            return existing_logger
    
    # Create new logger as either:
    # - No existing logger was found
    # or
    # - Existing logger was a null logger
    logger = logging.getLogger(module_name)  
    logger.setLevel(level)
    logger.propagate = False

    # Clear any existing handlers (important for converting null logger to real logger)
    logger.handlers.clear()

    formatter = ColorFormatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    handler = logging.StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    _configured_loggers[module_name] = logger
    return logger

def _get_null_logger(module_name: str) -> logging.Logger:
    """Get a null logger for the module

    Args:
        module_name (str): The name of the module to get the logger for.

    Returns:
        logging.Logger: The null logger for the module.
    """
    logger = logging.getLogger(module_name)
    logger.addHandler(logging.NullHandler())
    logger.propagate = False
    _configured_loggers[module_name] = logger
    return logger

def get_logger(
    module_name: str,
    level: int | None = WARNING,
    null_logger: bool | None = False,
) -> logging.Logger:
    """Get a logger for the module
    
    Validates the log level provided is a valid python logging log level.
    Returns a null logger if the null_logger argument is True,
    otherwise returns a real logger.
    
    Args:
        module_name (str): The name of the module to get the logger for.
        level (logging.Level | None): The log level to set for the logger.
            Defaults to logging.WARNING.
        null_logger (bool | None): Whether to return a null logger.
            Defaults to False.

    Returns:
        logging.Logger: The logger for the module.
    """
    if level is not None:
        _validate_log_level(level)
        
    if null_logger:
        return _get_null_logger(module_name=module_name)
    else:
        return _get_real_logger(module_name=module_name, level=level)



from __future__ import annotations

from typing import Final, TYPE_CHECKING

if TYPE_CHECKING:
    from ..enums import _BaseErrorReference

from ..messages import (
    INPUT_ERROR_MESSAGE,
    MAIN_CHAIN_JSON_ERROR_MESSAGE,
    FALLBACK_CHAIN_JSON_ERROR_MESSAGE,
    MAIN_CHAIN_VALIDATION_ERROR_MESSAGE,
    FALLBACK_CHAIN_VALIDATION_ERROR_MESSAGE,
    MAIN_CHAIN_VALUE_ERROR_MESSAGE,
    FALLBACK_CHAIN_VALUE_ERROR_MESSAGE,
    MAIN_CHAIN_TYPE_ERROR_MESSAGE,
    FALLBACK_CHAIN_TYPE_ERROR_MESSAGE,
    UNKNOWN_ERROR_MESSAGE,
    UNSUPPORTED_LLM_MODEL_ERROR_MESSAGE,
    INVALID_PARSER_TYPE_ERROR_MESSAGE,
    MISSING_PYDANTIC_MODEL_ERROR_MESSAGE,
    MISSING_OUTPUT_KEY_ERROR_MESSAGE,
    INVALID_PARSER_COMBINATION_ERROR_MESSAGE,
    MISSING_PROMPT_VARIABLES_ERROR_MESSAGE,
    INVALID_PROMPT_VARIABLES_ERROR_MESSAGE,
    INVALID_API_KEY_ERROR_MESSAGE,
)

from ..enums import (
    _ChainBuilderErrorReference,
    _ChainComposerErrorReference,
    _ChainWrapperErrorReference,
    _ChainManagerErrorReference,
    _ErrorType,
)

_ERROR_MESSAGE_MAP: Final[dict[_BaseErrorReference, dict[_ErrorType, str]]] = {
    _ChainWrapperErrorReference.INPUT_ERROR: {
        _ErrorType.VALUE_ERROR: INPUT_ERROR_MESSAGE,
    },
    
    _ChainWrapperErrorReference.MAIN_CHAIN_ERROR: {
        _ErrorType.JSON_ERROR: MAIN_CHAIN_JSON_ERROR_MESSAGE,
        _ErrorType.VALIDATION_ERROR: MAIN_CHAIN_VALIDATION_ERROR_MESSAGE,
        _ErrorType.VALUE_ERROR: MAIN_CHAIN_VALUE_ERROR_MESSAGE,
        _ErrorType.TYPE_ERROR: MAIN_CHAIN_TYPE_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
    
    _ChainWrapperErrorReference.FALLBACK_CHAIN_ERROR: {
        _ErrorType.JSON_ERROR: FALLBACK_CHAIN_JSON_ERROR_MESSAGE,
        _ErrorType.VALIDATION_ERROR: FALLBACK_CHAIN_VALIDATION_ERROR_MESSAGE,
        _ErrorType.VALUE_ERROR: FALLBACK_CHAIN_VALUE_ERROR_MESSAGE,
        _ErrorType.TYPE_ERROR: FALLBACK_CHAIN_TYPE_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
    
    _ChainComposerErrorReference.UNSUPPORTED_LLM_MODEL: {
        _ErrorType.VALUE_ERROR: UNSUPPORTED_LLM_MODEL_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
    
    _ChainComposerErrorReference.INVALID_PARSER_TYPE: {
        _ErrorType.VALUE_ERROR: INVALID_PARSER_TYPE_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
    
    _ChainComposerErrorReference.MISSING_PYDANTIC_MODEL: {
        _ErrorType.VALUE_ERROR: MISSING_PYDANTIC_MODEL_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
    
    _ChainComposerErrorReference.MISSING_OUTPUT_KEY: {
        _ErrorType.VALUE_ERROR: MISSING_OUTPUT_KEY_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
    
    _ChainComposerErrorReference.INVALID_PARSER_COMBINATION: {
        _ErrorType.VALUE_ERROR: INVALID_PARSER_COMBINATION_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
    
    _ChainComposerErrorReference.MISSING_PROMPT_VARIABLES: {
        _ErrorType.VALUE_ERROR: MISSING_PROMPT_VARIABLES_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
    
    _ChainComposerErrorReference.INVALID_PROMPT_VARIABLES: {
        _ErrorType.VALUE_ERROR: INVALID_PROMPT_VARIABLES_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
    
    _ChainComposerErrorReference.INVALID_API_KEY: {
        _ErrorType.VALUE_ERROR: INVALID_API_KEY_ERROR_MESSAGE,
        _ErrorType.UNKNOWN_ERROR: UNKNOWN_ERROR_MESSAGE,
    },
}



import os
import pytest
from pydantic import BaseModel
from typing import Dict, Any
from unittest.mock import patch

from chain_composer import ChainComposer
from chain_composer._error_handling import _ChainError

# Test Models
class TestOutputModel(BaseModel):
    result: str

class TestInputModel(BaseModel):
    input: str

# Fixtures
@pytest.fixture
def api_key() -> str:
    """Get API key from environment"""
    api_key = os.getenv("OPENAI_API_KEY")
    if not api_key:
        pytest.skip("OPENAI_API_KEY not found in environment")
    return api_key

@pytest.fixture
def basic_composer(api_key: str) -> ChainComposer:
    """Create a basic ChainComposer instance"""
    return ChainComposer(
        model="gpt-4",
        api_key=api_key,
    )

@pytest.fixture
def system_prompt() -> str:
    return """You are a helpful assistant that formats responses as JSON:
    {{
        "result": "<your response>"
    }}
    """

@pytest.fixture
def second_layer_system_prompt() -> str:
    return """You are a helpful assistant that takes the previous result and formats it:
    {{
        "result": "<processed previous result>"
    }}
    
    Previous result: {first_result}
    """

@pytest.fixture
def human_prompt() -> str:
    return "Process this input: {input_text}"

# Basic Initialization Tests
def test_chain_composer_initialization(api_key: str):
    """Test basic initialization of ChainComposer"""
    composer = ChainComposer(
        model="gpt-4",
        api_key=api_key,
    )
    assert composer is not None
    assert composer.api_key == api_key
    assert composer.llm_model == "gpt-4"
    assert composer.llm_model_type == "openai"

def test_chain_composer_invalid_model():
    """Test initialization with invalid model"""
    with pytest.raises(_ChainError):
        ChainComposer(
            model="invalid-model",
            api_key="dummy-key",
        )

# Chain Layer Tests
def test_add_single_chain_layer(basic_composer: ChainComposer, system_prompt: str, human_prompt: str):
    """Test adding a single chain layer"""
    composer = basic_composer.add_chain_layer(
        system_prompt=system_prompt,
        human_prompt=human_prompt,
        output_passthrough_key_name="result",
        parser_type="json",
        pydantic_output_model=TestOutputModel
    )
    
    assert len(composer.get_chain_sequence()) == 1

def test_add_multiple_chain_layers(basic_composer: ChainComposer, system_prompt: str, human_prompt: str):
    """Test adding multiple chain layers"""
    composer = basic_composer.add_chain_layer(
        system_prompt=system_prompt,
        human_prompt=human_prompt,
        output_passthrough_key_name="first_result",
        parser_type="json",
        pydantic_output_model=TestOutputModel
    ).add_chain_layer(
        system_prompt=system_prompt,
        human_prompt=human_prompt,
        output_passthrough_key_name="second_result",
        parser_type="pydantic",
        pydantic_output_model=TestOutputModel
    )
    
    assert len(composer.get_chain_sequence()) == 2

# Execution Tests
@patch('langchain_openai.ChatOpenAI')
def test_simple_chain_execution(mock_openai, basic_composer: ChainComposer, system_prompt: str, human_prompt: str, mock_openai_response):
    """Test executing a simple chain"""
    # Configure mock
    mock_instance = mock_openai.return_value
    mock_instance.invoke.return_value = mock_openai_response
    
    composer = basic_composer.add_chain_layer(
        system_prompt=system_prompt,
        human_prompt=human_prompt,
        output_passthrough_key_name="result",
        parser_type="json",
        pydantic_output_model=TestOutputModel
    )
    
    result = composer.run(
        prompt_variables_dict={
            "input_text": "Hello, world!"
        }
    )
    
    assert "result" in result
    assert isinstance(result["result"], dict)
    assert "result" in result["result"]

@patch('langchain_openai.ChatOpenAI')
def test_multi_layer_chain_execution(mock_openai, basic_composer: ChainComposer, system_prompt: str, human_prompt: str, second_layer_system_prompt: str, mock_openai_response):
    """Test executing a multi-layer chain"""
    # Configure mock
    mock_instance = mock_openai.return_value
    mock_instance.invoke.return_value = mock_openai_response
    
    composer = basic_composer.add_chain_layer(
        system_prompt=system_prompt,
        human_prompt=human_prompt,
        output_passthrough_key_name="first_result",
        parser_type="json",
        pydantic_output_model=TestOutputModel
    ).add_chain_layer(
        system_prompt=second_layer_system_prompt,
        human_prompt=human_prompt,
        output_passthrough_key_name="final_result",
        parser_type="pydantic",
        pydantic_output_model=TestOutputModel
    )
    
    result = composer.run(
        prompt_variables_dict={
            "input_text": "Test input"
        }
    )
    
    assert "first_result" in result
    assert "final_result" in result
    assert isinstance(result["first_result"], dict)
    assert isinstance(result["final_result"], TestOutputModel)

# Error Handling Tests
def test_missing_api_key():
    """Test initialization without API key"""
    with pytest.raises(_ChainError):
        ChainComposer(
            model="gpt-4",
            api_key="",
        )

def test_invalid_parser_type(basic_composer: ChainComposer, system_prompt: str, human_prompt: str):
    """Test adding chain layer with invalid parser type"""
    with pytest.raises(_ChainError):
        basic_composer.add_chain_layer(
            system_prompt=system_prompt,
            human_prompt=human_prompt,
            output_passthrough_key_name="result",
            parser_type="invalid",  # type: ignore
            pydantic_output_model=TestOutputModel
        )

def test_missing_pydantic_model(basic_composer: ChainComposer, system_prompt: str, human_prompt: str):
    """Test adding chain layer with pydantic parser but no model"""
    with pytest.raises(_ChainError):
        basic_composer.add_chain_layer(
            system_prompt=system_prompt,
            human_prompt=human_prompt,
            output_passthrough_key_name="result",
            parser_type="pydantic",
        )

# Chain Variables Tests
def test_chain_variables_update(basic_composer: ChainComposer):
    """Test updating chain variables"""
    variables = {"test_key": "test_value"}
    basic_composer._update_chain_variables(variables)
    assert basic_composer.get_chain_variables() == variables 


from __future__ import annotations

from typing import Type, Dict, TypeVar, TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from ..classes import _ChainError

class _ErrorRegistry:
    """Registry for mapping exception types to their handlers."""
    
    _registry: Dict[Type[Exception], Type[_ChainError]] = {}
    
    @classmethod
    def register(
        cls,
        handled_exception: Type[Exception],
    ) -> Callable[[Type[_ChainError]], Type[_ChainError]]:
        """Decorator to register an error handler for a specific exception type.
        
        Args:
            handled_exception: The exception type this handler can process
            
        Returns:
            Callable[[Type[T]], Type[T]]: A decorator that registers the handler for the exception type.
        """
        def decorator(handler_cls: Type[_ChainError]) -> Type[_ChainError]:
            cls._registry[handled_exception] = handler_cls
            return handler_cls
        return decorator
    
    @classmethod
    def get_handler(
        cls,
        exception: Exception,
    ) -> Type[_ChainError] | None:
        """Get the appropriate handler for an exception type.
        
        Args:
            exception: The exception to get a handler for.
            
        Returns:
            The handler for the exception type, or None if no handler is found.
        """
        for exc_type, handler in cls._registry.items():
            if isinstance(exception, exc_type):
                return handler
        return None


from .type_aliases import (
    FirstCallRequired,
    ParserUnion,
    ParserType,
    FallbackParserType,
    LLMUnion,
)

__all__ = [
    "FirstCallRequired",
    "ParserUnion",
    "ParserType",
    "FallbackParserType",
    "LLMUnion",
]



import json
from pydantic import ValidationError
from ..enums._error_types import _ErrorType

def get_error_type(error: Exception) -> _ErrorType:
    """Get the error type from an exception.

    Args:
        error (Exception): The exception to get the error type from.

    Returns:
        _ErrorType: The error type.
    """
    if isinstance(error, json.JSONDecodeError):
        return _ErrorType.JSON_ERROR
    elif isinstance(error, ValidationError):
        return _ErrorType.VALIDATION_ERROR
    elif isinstance(error, ValueError):
        return _ErrorType.VALUE_ERROR
    elif isinstance(error, TypeError):
        return _ErrorType.TYPE_ERROR
    else:
        return _ErrorType.UNKNOWN_ERROR


from typing import Union

from ..registry import _ErrorRegistry
from ..enums import _BaseErrorReference
from ..classes import _ChainError

class _ChainExceptionFactory:
    """Factory class for creating chain-specific exceptions."""
    
    @staticmethod
    def create_error(
        error: Exception,
        error_reference: _BaseErrorReference | None = None,
        message: str | None = None,
    ) -> _ChainError:
        """Creates appropriate chain exception based on the error type.

        Args:
            error (Exception): The original error that occurred
            error_reference (_BaseErrorReference | None): Reference for looking up error messages
            message (str | None): Optional custom error message
            
        Returns:
            _ChainError: The appropriate chain exception type
        """
        handler_cls = _ErrorRegistry.get_handler(error)
        
        if handler_cls is not None:
            return handler_cls(
                original_error=error,
                error_reference=error_reference
            )
            
        return _ChainError(
            message=message,
            original_error=error,
            error_reference=error_reference
        )



from typing import Final

UNSUPPORTED_LLM_MODEL_ERROR_MESSAGE: Final[str] = "Unsupported LLM model: {error}"
INVALID_PARSER_TYPE_ERROR_MESSAGE: Final[str] = "Invalid parser type: {error}"
MISSING_PYDANTIC_MODEL_ERROR_MESSAGE: Final[str] = "Missing Pydantic model for parser: {error}"
MISSING_OUTPUT_KEY_ERROR_MESSAGE: Final[str] = "Missing output key: {error}"
INVALID_PARSER_COMBINATION_ERROR_MESSAGE: Final[str] = "Invalid parser combination: {error}"
MISSING_PROMPT_VARIABLES_ERROR_MESSAGE: Final[str] = "Missing prompt variables: {error}"
INVALID_PROMPT_VARIABLES_ERROR_MESSAGE: Final[str] = "Invalid prompt variables: {error}"
INVALID_API_KEY_ERROR_MESSAGE: Final[str] = "Invalid API key: {error}"



from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, Optional
import warnings

from langchain.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    PromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain_openai import ChatOpenAI
from langchain_anthropic import ChatAnthropic
from langchain_google_genai import ChatGoogleGenerativeAI

from ..types import ValidationResult

if TYPE_CHECKING:
    from langchain.schema.runnable import Runnable

class APIKeyValidator:
    """Validator for LLM API keys across different services.

    This class validates API keys for different LLM services like OpenAI, Anthropic,
    and Google. It caches validation results to avoid redundant validation checks.

    Attributes:
        _validated_already (Dict[str, ValidationResult]): A dictionary that stores
            validation results for each API key, where the key is the API key string
            and the value is a ValidationResult object containing validation status
            for each service.

    Examples:
        Basic usage of the validator:

            >>> validator = APIKeyValidator()
            >>> if validator.is_valid(api_key="sk-...", model="gpt-4"):
            ...     print("Key is valid!")

        The validator caches results for subsequent checks:

            >>> # First check validates against services
            >>> validator.is_valid("sk-...", "gpt-4")
            >>> # Second check uses cached result
            >>> validator.is_valid("sk-...", "gpt-4")
    """

    def __init__(self):
        """Initialize the APIKeyValidator class."""
        # Dict to track api keys which have been validated already
        # Key = api_key, Value = ValidationResult
        self._validated_already: Dict[str, ValidationResult] = {}
        
    def is_valid(self, api_key: str, model: Optional[str] = None) -> bool:
        """Check if the API key is valid for any service. Validates if not already done."""
        if api_key not in self._validated_already:
            self._validate(api_key=api_key, model=model)

        results = self._validated_already[api_key]
        return any([results.openai, results.anthropic, results.google])
    
    def get_results_by_key_as_object(self, api_key: str) -> ValidationResult:
        """Get detailed validation results. Validates if not already done.
        
        Returns:
            ValidationResult: The validation results.
        """
        if api_key not in self._validated_already:
            raise ValueError(
                f"API key {self._mask(api_key)} has not been validated yet. "
                "Please validate the API key before calling this method.\n"
                "You can do so by calling the `is_valid` method on an APIKeyValidator instance.\n"
                ">>> validator = APIKeyValidator()\n"
                ">>> is_valid: bool = validator.is_valid(api_key, model)\n\n"
                "Then you can call the `get_results_by_key_as_object` method to recieve a ValidationResult object. "
                "This object is a dataclass with the following attributes:\n"
                "openai: bool\n"
                "anthropic: bool\n"
                "google: bool\n\n"
                "Usage Example:\n"
                ">>> results: ValidationResult = validator.get_results_by_key_as_object(api_key)\n"
                ">>> print(results.openai)\n"
                ">>> print(results.anthropic)\n"
                ">>> print(results.google)\n\n"
            )

        return self._validated_already[api_key]

    def get_results_for_api_key(self, api_key: str) -> Dict[str, bool]:
        """Get detailed validation results. Validates if not already done."""
        if api_key not in self._validated_already:
            raise ValueError(
                f"API key {self._mask(api_key)} has not been validated yet. "
                "Please validate the API key before calling this method.\n"
                "You can do so by calling the `is_valid` method on an APIKeyValidator instance.\n"
                ">>> validator = APIKeyValidator()\n"
                ">>> is_valid: bool = validator.is_valid(api_key, model)\n\n"
                "Then you can call the `get_results_for_api_key` method to recieve a dictionary of results. "
                "This dictionary has the following keys:\n"
                "openai: bool\n"
                "anthropic: bool\n"
                "google: bool\n\n"
                "Usage Example:\n"
                ">>> results: Dict[str, bool] = validator.get_results_for_api_key(api_key)\n"
                ">>> print(results['openai'])\n"
                ">>> print(results['anthropic'])\n"
                ">>> print(results['google'])\n\n"
            )
        
        results = self._validated_already[api_key]
        
        return {
            "openai": results.openai,
            "anthropic": results.anthropic,
            "google": results.google,
        }

    def _validate(
        self, 
        api_key: str, 
        model: Optional[str] = None,
    ) -> None:
        """Run validation tests for each service."""
        
        results: ValidationResult = ValidationResult()
        prompt: ChatPromptTemplate = self._get_test_prompt()
        
        # Store warnings instead of immediately showing them
        warnings_list = []
        
        def test_with_warning_capture(provider_name: str, provider_func) -> bool:
            result, warning = self._test_provider(
                api_key,
                provider_func(api_key, model),
                prompt,
                model,
            )
            if not result and warning:
                warnings_list.append(f"{provider_name}: {warning}")
            return result
            
        # Test each provider
        results.openai = test_with_warning_capture("OpenAI", self._get_openai_llm)
        results.anthropic = test_with_warning_capture("Anthropic", self._get_anthropic_llm)
        results.google = test_with_warning_capture("Google", self._get_google_llm)
        
        # Only show warnings if all providers failed
        if not any([results.openai, results.anthropic, results.google]):
            for warning in warnings_list:
                warnings.warn(warning)
        
        self._validated_already[api_key] = results
        
    def _get_test_prompt(self) -> ChatPromptTemplate:
        """Get a test prompt for the API key validator.
        
        Returns:
            ChatPromptTemplate: The test prompt.
        """
        system_prompt_template: PromptTemplate = PromptTemplate(template="test")
        human_prompt_template: PromptTemplate = PromptTemplate(template="test")

        prompt: ChatPromptTemplate = ChatPromptTemplate.from_messages(
            [
                SystemMessagePromptTemplate.from_template(
                    system_prompt_template.template
                ),
                HumanMessagePromptTemplate.from_template(
                    human_prompt_template.template
                ),
            ]
        )
        
        return prompt
    
    def _get_openai_llm(
        self, 
        api_key: str, 
        model: Optional[str] = None,
    ) -> ChatOpenAI:
        """Get an OpenAI LLM instance."""
        if model is None:
            return ChatOpenAI(api_key=api_key)
        else:
            return ChatOpenAI(api_key=api_key, model=model)
    
    def _get_anthropic_llm(
        self, 
        api_key: str, 
        model: Optional[str] = None,
    ) -> ChatAnthropic:
        """Get an Anthropic LLM instance."""
        if model is None:
            return ChatAnthropic(api_key=api_key)
        else:
            return ChatAnthropic(api_key=api_key, model=model)
    
    def _get_google_llm(
        self, 
        api_key: str, 
        model: Optional[str] = None,
    ) -> ChatGoogleGenerativeAI:
        """Get a Google Generative AI LLM instance."""
        if model is None:
            return ChatGoogleGenerativeAI(api_key=api_key)
        else:
            return ChatGoogleGenerativeAI(api_key=api_key, model=model)
    
    def _test_provider(
        self,
        api_key: str,
        llm: ChatOpenAI | ChatAnthropic | ChatGoogleGenerativeAI,
        prompt: ChatPromptTemplate,
        model: str | None = None,
    ) -> tuple[bool, str | None]:
        """Run a test to see if the API key is valid for a given provider.
        
        Args:
            api_key (str): The API key to test.
            llm (ChatOpenAI | ChatAnthropic | ChatGoogleGenerativeAI): The LLM to test.
            prompt (ChatPromptTemplate): The prompt to test the LLM with.
            model (str | None): The model to test the LLM with.

        Returns:
            tuple[bool, str | None]: A tuple containing a boolean indicating if the API key is valid, and a string containing a warning message if the API key is not valid.
        """
        try:
            chain: Runnable = prompt | llm
            chain.invoke({})
            return True, None
        except Exception as e:
            error_type = str(type(e).__name__)
            if any(err in error_type for err in [
                # OpenAI
                "AuthenticationError",
                "PermissionError",
                
                # Anthropic
                "AuthenticationError",
                
                # Google
                "PermissionDenied",
                "InvalidArgument",
                
                # General
                "RequestException",
            ]):
                warning_msg = (
                    f"The API key: {self._mask(api_key)} has an issue associated with it.\n"
                    f"Error: {str(e)}\n"
                    f"Provider: {type(llm)}\n"
                    f"Model: {model}\n"
                )
                return False, warning_msg
            else:
                warning_msg = f"Unexpected error during API key validation: {str(e)}"
                return False, warning_msg
        
    def _mask(self, api_key: str) -> str:
        """Mask the API key for display purposes."""
        return api_key[:4] + "..." + api_key[-4:]



from dataclasses import dataclass

@dataclass
class ValidationResult:
    """Class to store the validation results. Used by APIKeyValidator.
    
    Attributes:
        openai (bool): True if the OpenAI API key is valid, False otherwise.
        anthropic (bool): True if the Anthropic API key is valid, False otherwise.
        google (bool): True if the Google API key is valid, False otherwise.
    """
    openai: bool = False
    anthropic: bool = False
    google: bool = False


from .classes._llm_validation_error import _ChainError
from .enums._error_types import _ErrorType
from .enums._error_reference import _BaseErrorReference, _ChainComposerErrorReference, _ChainWrapperErrorReference, _ChainBuilderErrorReference, _ChainManagerErrorReference
from .factories import _ChainExceptionFactory
from .validators import _ChainComposerValidator

__all__ = [
    '_ChainError',
    '_ErrorType',
    '_BaseErrorReference',
    '_ChainExceptionFactory',
    '_ChainComposerValidator',
    '_ChainComposerErrorReference',
    '_ChainWrapperErrorReference',
    '_ChainBuilderErrorReference',
    '_ChainManagerErrorReference'
]


from ._error_reference import (
    _BaseErrorReference,
    _ChainBuilderErrorReference,
    _ChainComposerErrorReference,
    _ChainWrapperErrorReference,
    _ChainManagerErrorReference,
)

from ._error_types import (
    _ErrorType,
)

__all__ = [
    "_BaseErrorReference",
    "_ChainBuilderErrorReference",
    "_ChainComposerErrorReference",
    "_ChainWrapperErrorReference",
    "_ChainManagerErrorReference",
    "_ErrorType",
]



from ._chain_wrapper_error_messages import (
    INPUT_ERROR_MESSAGE,
    MAIN_CHAIN_JSON_ERROR_MESSAGE,
    FALLBACK_CHAIN_JSON_ERROR_MESSAGE,
    MAIN_CHAIN_VALIDATION_ERROR_MESSAGE,
    FALLBACK_CHAIN_VALIDATION_ERROR_MESSAGE,
    MAIN_CHAIN_VALUE_ERROR_MESSAGE,
    FALLBACK_CHAIN_VALUE_ERROR_MESSAGE,
    MAIN_CHAIN_TYPE_ERROR_MESSAGE,
    FALLBACK_CHAIN_TYPE_ERROR_MESSAGE,
)

from ._universal_error_messages import UNKNOWN_ERROR_MESSAGE

from ._chain_composer_error_messages import (
    UNSUPPORTED_LLM_MODEL_ERROR_MESSAGE,
    INVALID_PARSER_TYPE_ERROR_MESSAGE,
    MISSING_PYDANTIC_MODEL_ERROR_MESSAGE,
    MISSING_OUTPUT_KEY_ERROR_MESSAGE,
    INVALID_PARSER_COMBINATION_ERROR_MESSAGE,
    MISSING_PROMPT_VARIABLES_ERROR_MESSAGE,
    INVALID_PROMPT_VARIABLES_ERROR_MESSAGE,
    INVALID_API_KEY_ERROR_MESSAGE,
)

__all__ = [
    "INPUT_ERROR_MESSAGE",
    "MAIN_CHAIN_JSON_ERROR_MESSAGE",
    "FALLBACK_CHAIN_JSON_ERROR_MESSAGE",
    "MAIN_CHAIN_VALIDATION_ERROR_MESSAGE",
    "FALLBACK_CHAIN_VALIDATION_ERROR_MESSAGE",
    "MAIN_CHAIN_VALUE_ERROR_MESSAGE",
    "FALLBACK_CHAIN_VALUE_ERROR_MESSAGE",
    "MAIN_CHAIN_TYPE_ERROR_MESSAGE",
    "FALLBACK_CHAIN_TYPE_ERROR_MESSAGE",
    "UNKNOWN_ERROR_MESSAGE",
    "UNSUPPORTED_LLM_MODEL_ERROR_MESSAGE",
    "INVALID_PARSER_TYPE_ERROR_MESSAGE",
    "MISSING_PYDANTIC_MODEL_ERROR_MESSAGE",
    "MISSING_OUTPUT_KEY_ERROR_MESSAGE",
    "INVALID_PARSER_COMBINATION_ERROR_MESSAGE",
    "MISSING_PROMPT_VARIABLES_ERROR_MESSAGE",
    "INVALID_PROMPT_VARIABLES_ERROR_MESSAGE",
    "INVALID_API_KEY_ERROR_MESSAGE",
]


import pytest
from langchain.prompts import ChatPromptTemplate
from langchain.schema.runnable import Runnable
from langchain_core.output_parsers import PydanticOutputParser, StrOutputParser
from langchain_openai import ChatOpenAI
from pydantic import BaseModel

from chain_composer._core_internal._chain_builder import _ChainBuilder
from chain_composer._logging import WARNING

class TestOutputModel(BaseModel):
    response: str

@pytest.fixture
def chat_prompt():
    return ChatPromptTemplate.from_messages([
        ("system", "You are a helpful assistant."),
        ("human", "{input}")
    ])

@pytest.fixture
def llm():
    return ChatOpenAI(model="gpt-3.5-turbo", temperature=0)

@pytest.fixture
def pydantic_parser():
    return PydanticOutputParser(pydantic_object=TestOutputModel)

@pytest.fixture
def str_parser():
    return StrOutputParser()

@pytest.fixture
def basic_builder(chat_prompt, llm):
    return _ChainBuilder(
        chat_prompt=chat_prompt,
        llm=llm,
        enable_logging=True,
        level=WARNING
    )

@pytest.fixture
def builder_with_parser(chat_prompt, llm, pydantic_parser):
    return _ChainBuilder(
        chat_prompt=chat_prompt,
        llm=llm,
        parser=pydantic_parser,
        enable_logging=True,
        level=WARNING
    )

@pytest.fixture
def builder_with_fallback(chat_prompt, llm, pydantic_parser, str_parser):
    return _ChainBuilder(
        chat_prompt=chat_prompt,
        llm=llm,
        parser=pydantic_parser,
        fallback_parser=str_parser,
        enable_logging=True,
        level=WARNING
    )

def test_chain_builder_initialization(basic_builder):
    """Test basic initialization of ChainBuilder"""
    assert basic_builder.chat_prompt is not None
    assert basic_builder.llm is not None
    assert basic_builder.parser is None
    assert basic_builder.fallback_parser is None
    assert basic_builder.debug is False
    assert isinstance(basic_builder.chain, Runnable)
    assert isinstance(basic_builder.fallback_chain, Runnable)

def test_chain_builder_with_parser(builder_with_parser):
    """Test ChainBuilder initialization with parser"""
    assert builder_with_parser.parser is not None
    assert isinstance(builder_with_parser.parser, PydanticOutputParser)
    assert builder_with_parser.fallback_parser is None

def test_chain_builder_with_fallback(builder_with_fallback):
    """Test ChainBuilder initialization with both parsers"""
    assert builder_with_fallback.parser is not None
    assert builder_with_fallback.fallback_parser is not None
    assert isinstance(builder_with_fallback.parser, PydanticOutputParser)
    assert isinstance(builder_with_fallback.fallback_parser, StrOutputParser)

def test_chain_builder_debug_warning():
    """Test debug warning when None is provided"""
    with pytest.warns(UserWarning, match="The `debug` argument takes an optional boolean value"):
        _ChainBuilder(
            chat_prompt=ChatPromptTemplate.from_messages([("system", "test")]),
            llm=ChatOpenAI(),
            debug=None
        )

def test_chain_builder_str_representation(basic_builder):
    """Test string representation of ChainBuilder"""
    str_repr = str(basic_builder)
    assert "ChainBuilder" in str_repr
    assert "ChatPromptTemplate" in str_repr
    assert "ChatOpenAI" in str_repr

def test_chain_builder_repr_representation(basic_builder):
    """Test detailed representation of ChainBuilder"""
    repr_str = repr(basic_builder)
    assert "ChainBuilder" in repr_str
    assert "chat_prompt" in repr_str
    assert "llm" in repr_str
    assert "parser" in repr_str
    assert "fallback_parser" in repr_str

def test_get_chain(basic_builder):
    """Test get_chain method returns a Runnable"""
    chain = basic_builder.get_chain()
    assert isinstance(chain, Runnable)

def test_get_fallback_chain(basic_builder):
    """Test get_fallback_chain method returns a Runnable"""
    fallback_chain = basic_builder.get_fallback_chain()
    assert isinstance(fallback_chain, Runnable) 


from __future__ import annotations

from typing import TYPE_CHECKING
from ..utils import _ERROR_MESSAGE_MAP
from ..enums import _BaseErrorReference, _ErrorType

if TYPE_CHECKING:
    from ..enums import _BaseErrorReference

class _ChainError(Exception):
    """Base exception for all chain-related errors."""
    
    _CLASS_ERROR_TYPE = _ErrorType.UNKNOWN_ERROR
    _INVALID_STATE_ERROR_MESSAGE = ""
    
    def __init__(
        self,
        message: str | None = None,
        error_reference: _BaseErrorReference | None = None,
        original_error: Exception | None = None,
        class_error_type: _ErrorType | None = None,
    ) -> None:
        """Initialize the ChainError.

        Args:
            message (str): The error message.
            original_error (Exception | None, optional): The original error that occurred. Defaults to None.

        Raises:
            ValueError: If original_error is provided but is not an instance of Exception.
        """
        self.message = message
        self.original_error = original_error
        self.error_reference = error_reference
        
        if class_error_type is not None:
            self.set_class_error_type(class_error_type)
        
        if not self._is_valid_state(message, original_error, error_reference):
            raise ValueError(self._INVALID_STATE_ERROR_MESSAGE)
        
        message = message if message is not None else ""
        
        if error_reference is not None:
            self._raise_error_with_ref(original_error, error_reference)
        else:
            self._raise_error(message, original_error)
    
    def set_class_error_type(
        self, 
        error_type: _ErrorType,
    ) -> None:
        """Set the class error type.
        
        Args:
            error_type (ErrorType): The error type.
        """
        self._CLASS_ERROR_TYPE = error_type

    def _raise_error(
        self, 
        message: str, 
        original_error: Exception | None,
    ) -> None:
        """Initialize the parent Exception class with appropriate arguments.
        
        Args:
            message (str): The error message.
            original_error (Exception | None): The original error that occurred.
        """
        if self._original_error_exists(original_error):
            super().__init__(message, original_error)
        else:
            super().__init__(message)
    
    def _raise_error_with_ref(
        self,
        original_error: Exception | None,
        error_reference: _BaseErrorReference,
    ) -> None:
        """Raise the error with the error reference.
        
        Args:
            original_error (Exception | None): The original error that occurred.
            error_reference (BaseErrorReference | None): The error reference that occurred.
        """
        message = _ERROR_MESSAGE_MAP[error_reference][self._CLASS_ERROR_TYPE]
        
        if self._original_error_exists(original_error):
            super().__init__(message, original_error)
        else:
            super().__init__(message)

    def _is_valid_state(
        self, 
        message: str | None, 
        original_error: Exception | None,
        error_reference: _BaseErrorReference | None,
    ) -> bool:
        """Check if the original error is in a valid state.
        
        Args:
            message (str | None): The message that occurred.
            original_error (Exception | None): The original error that occurred.
            error_reference (BaseErrorReference | None): The error reference that occurred.

        Returns:
            bool: True if the original error is in a valid state, False otherwise
        """
        is_valid: bool = False
        message_exists: bool = self._message_exists(message)
        error_reference_exists: bool = self._error_reference_exists(error_reference)
        original_error_exists: bool = self._original_error_exists(original_error)
        original_error_is_exception: bool = self._is_exception(original_error)
        
        if original_error_exists and not original_error_is_exception:
            self._INVALID_STATE_ERROR_MESSAGE += (
                "original_error must be an instance of Exception, "
                f"but is of type: {type(original_error)}\n\n"
            )
        
        if not message_exists and not original_error_exists:
            self._INVALID_STATE_ERROR_MESSAGE += (
                "A message or original error must be provided\n\n"
            )
            
        if message_exists and error_reference_exists:
            self._INVALID_STATE_ERROR_MESSAGE += (
                "Error references are for looking up error messages, "
                "so a message should not be provided. "
                "Instead, edit your error message associated with the error reference.\n\n"
            )
        
        if message_exists or (original_error_exists and original_error_is_exception):
            is_valid = True
        
        return is_valid
        
    def _error_reference_exists(
        self, 
        error_reference: _BaseErrorReference | None
    ) -> bool:
        return error_reference is not None
    
    def _message_exists(
        self, 
        message: str | None
    ) -> bool:
        """Check if the message is None.
        
        Args:
            message (str | None): The message that occurred.

        Returns:
            bool: True if the message is None, False otherwise
        """
        return message is not None

    def _original_error_exists(
        self, 
        original_error: Exception | None
    ) -> bool:
        """Check if the original error exists.
        
        Args:
            original_error (Exception | None): The original error that occurred.

        Returns:
            bool: True if the original error exists, False otherwise
        """
        return original_error is not None
    
    def _is_exception(
        self, 
        original_error: Exception | None
    ) -> bool:
        """Check if the original error is an instance of Exception.
        
        Args:
            original_error (Exception | None): The original error that occurred.

        Returns:
            bool: True if the original error is an instance of Exception, False otherwise
 
        Note:
            This method is used to check if the original error is an instance of Exception. It is used in the `__init__` method of the `ChainError` class.
        """
        return isinstance(original_error, Exception)


import json
from typing import Dict, Any

import pytest
from langchain.schema.runnable import Runnable, RunnablePassthrough
from langchain_core.output_parsers import PydanticOutputParser, StrOutputParser
from pydantic import BaseModel, ValidationError

from chain_composer._core_internal._chain_wrapper import _ChainWrapper
from chain_composer._error_handling import _ChainExceptionFactory
from chain_composer._logging import WARNING

class TestOutputModel(BaseModel):
    response: str

class MockRunnable(Runnable):
    def __init__(self, output=None, error=None):
        self.output = output
        self.error = error

    def invoke(self, input_data: Dict[str, Any], **kwargs):
        if self.error:
            raise self.error
        return self.output

@pytest.fixture
def mock_chain():
    return MockRunnable(output={"response": "test response"})

@pytest.fixture
def mock_fallback_chain():
    return MockRunnable(output="fallback response")

@pytest.fixture
def pydantic_parser():
    return PydanticOutputParser(pydantic_object=TestOutputModel)

@pytest.fixture
def str_parser():
    return StrOutputParser()

@pytest.fixture
def basic_wrapper(mock_chain, mock_fallback_chain):
    return _ChainWrapper(
        chain=mock_chain,
        fallback_chain=mock_fallback_chain,
        enable_logging=True,
        level=WARNING
    )

@pytest.fixture
def wrapper_with_parser(mock_chain, mock_fallback_chain, pydantic_parser, str_parser):
    return _ChainWrapper(
        chain=mock_chain,
        fallback_chain=mock_fallback_chain,
        parser=pydantic_parser,
        fallback_parser=str_parser,
        enable_logging=True,
        level=WARNING
    )

def test_chain_wrapper_initialization(basic_wrapper):
    """Test basic initialization of ChainWrapper"""
    assert basic_wrapper.chain is not None
    assert basic_wrapper.fallback_chain is not None
    assert basic_wrapper.parser is None
    assert basic_wrapper.fallback_parser is None
    assert basic_wrapper.preprocessor is None
    assert basic_wrapper.postprocessor is None
    assert basic_wrapper.debug is False

def test_chain_wrapper_with_parsers(wrapper_with_parser):
    """Test ChainWrapper initialization with parsers"""
    assert wrapper_with_parser.parser is not None
    assert wrapper_with_parser.fallback_parser is not None
    assert isinstance(wrapper_with_parser.parser, PydanticOutputParser)
    assert isinstance(wrapper_with_parser.fallback_parser, StrOutputParser)

def test_chain_wrapper_debug_warning():
    """Test debug warning when None is provided"""
    with pytest.warns(UserWarning, match="The `debug` argument takes an optional boolean value"):
        _ChainWrapper(
            chain=MockRunnable(),
            fallback_chain=MockRunnable(),
            debug=None
        )

def test_chain_wrapper_str_representation(basic_wrapper):
    """Test string representation of ChainWrapper"""
    str_repr = str(basic_wrapper)
    assert "ChainWrapper" in str_repr
    assert "parser=None" in str_repr
    assert "fallback_parser=None" in str_repr

def test_chain_wrapper_repr_representation(basic_wrapper):
    """Test detailed representation of ChainWrapper"""
    repr_str = repr(basic_wrapper)
    assert "ChainWrapper" in repr_str
    assert "chain" in repr_str
    assert "parser" in repr_str
    assert "fallback_parser" in repr_str
    assert "preprocessor" in repr_str
    assert "postprocessor" in repr_str

def test_run_chain_success(basic_wrapper):
    """Test successful chain execution"""
    result = basic_wrapper.run_chain(input_data={"input": "test"})
    assert result == {"response": "test response"}

def test_run_chain_with_preprocessor():
    """Test chain execution with preprocessor"""
    def preprocessor(data: Dict[str, Any]) -> Dict[str, Any]:
        data["processed"] = True
        return data

    wrapper = _ChainWrapper(
        chain=MockRunnable(output="test"),
        fallback_chain=MockRunnable(),
        preprocessor=preprocessor
    )
    result = wrapper.run_chain(input_data={"input": "test"})
    assert result == "test"

def test_run_chain_with_postprocessor():
    """Test chain execution with postprocessor"""
    def postprocessor(data: Any) -> str:
        return f"Processed: {data}"

    wrapper = _ChainWrapper(
        chain=MockRunnable(output="test"),
        fallback_chain=MockRunnable(),
        postprocessor=postprocessor
    )
    result = wrapper.run_chain(input_data={"input": "test"})
    assert result == "Processed: test"

def test_run_chain_json_error_with_fallback(mock_fallback_chain):
    """Test chain execution with JSON error and fallback"""
    error_chain = MockRunnable(error=json.JSONDecodeError("test", "test", 0))
    wrapper = _ChainWrapper(
        chain=error_chain,
        fallback_chain=mock_fallback_chain
    )
    result = wrapper.run_chain(input_data={"input": "test"})
    assert result == "fallback response"

def test_run_chain_validation_error_with_fallback(mock_fallback_chain):
    """Test chain execution with validation error and fallback"""
    error_chain = MockRunnable(error=ValidationError.from_exception_data("test", []))
    wrapper = _ChainWrapper(
        chain=error_chain,
        fallback_chain=mock_fallback_chain
    )
    result = wrapper.run_chain(input_data={"input": "test"})
    assert result == "fallback response"

def test_run_chain_error_without_fallback():
    """Test chain execution with error and no fallback"""
    error_chain = MockRunnable(error=ValueError("test error"))
    wrapper = _ChainWrapper(
        chain=error_chain,
        fallback_chain=MockRunnable()
    )
    with pytest.raises(Exception):
        wrapper.run_chain(input_data={"input": "test"}, is_last_chain=True)

def test_get_parser_type(wrapper_with_parser):
    """Test get_parser_type method"""
    assert wrapper_with_parser.get_parser_type() == "PydanticOutputParser"

def test_get_fallback_parser_type(wrapper_with_parser):
    """Test get_fallback_parser_type method"""
    assert wrapper_with_parser.get_fallback_parser_type() == "StrOutputParser"

def test_run_chain_with_pydantic_output():
    """Test chain execution with Pydantic model output"""
    output_model = TestOutputModel(response="test")
    chain = MockRunnable(output=output_model)
    wrapper = _ChainWrapper(
        chain=chain,
        fallback_chain=MockRunnable()
    )
    result = wrapper.run_chain(input_data={"input": "test"}, is_last_chain=False)
    assert isinstance(result, dict)
    assert result["response"] == "test" 


from __future__ import annotations

import logging
import warnings
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    TYPE_CHECKING
)

from .._logging import get_logger, WARNING

if TYPE_CHECKING:
    from pydantic import BaseModel
    from ._chain_wrapper import _ChainWrapper

class _ChainManager:
    """Manages a sequence of chain operations.

    ChainManager is a class that manages a sequence of chain operations, allowing for
    the addition of chains and running them in sequence with provided data.

    Methods:
        __init__(logger: Logger | None):
            Initializes the ChainManager instance with an optional logger.
            Type: logging.Logger | None
            Defaults to None.

        __str__() -> str:
            Returns a string representation of the ChainManager instance.

        __repr__() -> str:
            Returns a string representation of the ChainManager instance.

        add_chain(chain_wrapper, output_passthrough_key_name):
            Adds a chain to the chain sequence.

            Args:
                chain_wrapper (ChainWrapper): The chain wrapper to add
                output_passthrough_key_name (str | None): Optional output key name
                    Defaults to None.

        run(data_dict, data_dict_update_function):
            Runs the chain sequence with the provided data dictionary and optional update function.

            Args:
                data_dict (dict): The input data dictionary
                    Type: Dict[str, Any]
                data_dict_update_function (callable | None): Optional update function
                    Type: Callable[[Dict[str, Any]], None]
                    Defaults to None.
            Returns:
                dict: The updated data dictionary
                    Type: Dict[str, Any]
    """

    def __init__(
        self,
        enable_logging: bool | None = False,
        level: int | None = WARNING,
    ) -> None:
        """Initializes the ChainManager instance.

        Args:
            logger (Logger | None): A logger instance to be used for logging.
                Type: logging.Logger | None
                If not provided, a default logger will be created.
                Defaults to None.

        Attributes:
            logger (Logger): The logger instance used for logging.
                Type: logging.Logger
            chain_sequence (list): A list to store the chain sequence.
                Type: List[Tuple[_ChainWrapper, str | None]]
        """
        self.logger = get_logger(
            module_name=__name__,
            level=level,
            null_logger=not enable_logging,
        )

        self.chain_sequence: List[Tuple[_ChainWrapper, Optional[str]]] = []

    def __str__(self) -> str:
        """Returns a string representation of the ChainManager object.

        The string representation includes the index and the wrapper of each
        element in the chain_sequence.

        Returns:
            str: A string representation of the ChainManager object.
        """
        chain_info = ", ".join(
            [
                f"{idx}: {wrapper}"
                for idx, (wrapper, _) in enumerate(self.chain_sequence)
            ]
        )
        return f"ChainManager(chains=[{chain_info}])"

    def __repr__(self) -> str:
        """Returns a string representation of the object for debugging purposes.

        This method calls the __str__() method to provide a human-readable
        representation of the object.

        Returns:
            str: A string representation of the object.
        """
        return self.__str__()

    def add_chain(
        self,
        *,
        chain_wrapper: _ChainWrapper,
        output_passthrough_key_name: str | None = None,
    ) -> None:
        """Adds a chain to the chain sequence.

        Args:
            chain_wrapper (ChainWrapper): The chain wrapper to be added.
            output_passthrough_key_name (str | None): The key name for output passthrough.
                Type: str | None
                Defaults to None.

        Returns:
            None
        """
        self.chain_sequence.append((chain_wrapper, output_passthrough_key_name))

    def run(
        self,
        *,
        data_dict: Dict[str, Any],
        data_dict_update_function: Callable[[Dict[str, Any]], None] | None = None,
    ) -> Dict[str, Any]:
        """Executes a sequence of chains, updating the provided data dictionary with the results of each chain.

        Args:
            data_dict (dict): The initial data dictionary containing input variables for the chains.
                Type: Dict[str, Any]
            data_dict_update_function (callable | None): An optional function to update the data
                dictionary after each chain execution.
                Type: Callable[[Dict[str, Any]], None]
                Defaults to None.

        Returns:
            dict: The updated data dictionary after all chains have been executed.
                Type: Dict[str, Any]

        Raises:
            UserWarning: If the provided data dictionary is empty.

        Notes:
            - The method iterates over a sequence of chains (`self.chain_sequence`), executing
              each chain and updating the `data_dict` with the output.
            - If an `output_name` is provided for a chain, the output is stored in `data_dict`
              under that name; otherwise, it is stored under the key `_last_output`.
            - If `data_dict_update_function` is provided, it is called with the updated
              `data_dict` after each chain execution.
        """
        if not data_dict:
            warnings.warn(
                "No variables provided for the chain. Please ensure you have provided the necessary variables. If you have variable placeholders and do not pass them in it will result in an error."
            )

        num_chains: int = len(self.chain_sequence)
        for index, (chain_wrapper, output_name) in enumerate(self.chain_sequence):
            is_last_chain: bool = index == num_chains - 1
            output: Union[BaseModel, Dict[str, Any], str] = chain_wrapper.run_chain(
                input_data=data_dict, is_last_chain=is_last_chain
            )

            # Update data with the output
            if output_name:
                data_dict[output_name] = output
            else:
                data_dict["_last_output"] = output

            if data_dict_update_function:
                data_dict_update_function(data_dict)

        return data_dict


from ._chain_builder import _ChainBuilder
from ._chain_manager import _ChainManager
from ._chain_wrapper import _ChainWrapper

__all__ = [
    "_ChainBuilder",
    "_ChainManager",
    "_ChainWrapper",
]


from __future__ import annotations

import json
from typing import TYPE_CHECKING

from ._base_chain_error import _ChainError
from ..utils import get_error_type, _ERROR_MESSAGE_MAP
from ..enums import _BaseErrorReference, _ErrorType
from ..registry import _ErrorRegistry

if TYPE_CHECKING:
    from ..enums import _BaseErrorReference

@_ErrorRegistry.register(json.JSONDecodeError)
class _LLMJsonDecodeError(_ChainError):
    """Exception for JSON parsing failures in LLM output."""
    
    def __init__(
        self, 
        original_error: json.JSONDecodeError,
        error_reference: _BaseErrorReference | None = None,
    ) -> None:
        """Initialize the LLMJsonDecodeError.
        
        Args:
            original_error (json.JSONDecodeError): The original JSON decode error
            error_reference (BaseErrorReference | None): The error reference that occurred.
        """
        super().__init__(
            original_error=original_error,
            error_reference=error_reference,
            class_error_type=get_error_type(original_error),
        )


from __future__ import annotations

import warnings
from typing import (
    Optional,
    Union,
    TYPE_CHECKING,
    Type,
)

from langchain.prompts import ChatPromptTemplate
from langchain.schema.runnable import Runnable, RunnablePassthrough
from langchain_anthropic import ChatAnthropic
from langchain_core.output_parsers import PydanticOutputParser
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_openai import ChatOpenAI
from pydantic import BaseModel

from .._logging import get_logger, WARNING

if TYPE_CHECKING:
    from .._type_aliases import (
        ParserUnion,
        LLMUnion,
    )

class _ChainBuilder:
    """A builder class for constructing and configuring LangChain chains with logging and parsing capabilities.

    This class provides functionality to construct LangChain chains with customizable prompts,
    language models, and output parsers. It includes support for logging, fallback chains,
    and various parser types.

    Attributes:
        chat_prompt (ChatPromptTemplate): Template for chat prompts.
        llm (): Language model instance.

        parser (ParserType | None): Primary output parser.
        fallback_parser (ParserType | None): Fallback output parser.
        chain (:class:`Runnable`): Primary chain instance.
        fallback_chain (:class:`Runnable`): Fallback chain instance.
        debug (bool): Whether to enable debug logging.
    """

    def __init__(
        self,
        *,
        chat_prompt: ChatPromptTemplate,
        llm: LLMUnion,
        parser: ParserUnion | None = None,
        fallback_parser: ParserUnion | None = None,
        enable_logging: bool | None = False,
        level: int | None = WARNING,
        debug: bool | None = False,
    ) -> None:
        """Initialize a ChainBuilder instance.

        Args:
            chat_prompt (ChatPromptTemplate): Template for structuring chat interactions.
            llm (LLMUnion): Language model to use.
            parser (ParserType | None): Primary output parser.
                Defaults to None.
            fallback_parser (FallbackParserType | None): Fallback output parser.
                Defaults to None.
            enable_logging (bool | None): Whether to enable logging.
                Defaults to False.
            level (logging.Level | None): The log level to use for the module.
                Defaults to `logging.WARNING`.
            debug (bool | None): Whether to enable debug logging.
                Defaults to False.
        """
        self.logger = get_logger(
            module_name=__name__,
            level=level,
            null_logger=not enable_logging,
        )
        self.chat_prompt: ChatPromptTemplate = chat_prompt
        self.parser: Optional[ParserUnion] = parser
        self.fallback_parser: Optional[ParserUnion] = fallback_parser
        self.llm: Union[ChatOpenAI, ChatAnthropic, ChatGoogleGenerativeAI] = llm
        if debug is None:
            warnings.warn(
                "The `debug` argument takes an optional boolean value, you gave None. "
                "The value will be set to False.\n"
                "If your intention is to not enable debug logging, "
                "you can simply omit the `debug` argument.\n"
            )
        self.debug: bool = False if debug is None else debug
        self.chain: Runnable = self._build_chain()
        self.fallback_chain: Runnable = self._build_fallback_chain()

                
    def __str__(self) -> str:
        """Returns a concise string representation of the ChainBuilder object.
        
        Returns:
            str: A concise string representation of the ChainBuilder object.
        """
        return f"ChainBuilder(chat_prompt={type(self.chat_prompt).__name__}, llm={self.llm.__class__.__name__}, parser={type(self.parser).__name__ if self.parser else 'None'})"

    def __repr__(self) -> str:
        """Returns a detailed string representation for debugging.
        
        Returns:
            str: A detailed string representation of the ChainBuilder object.
        """
        return (
            f"ChainBuilder(\n"
            f"    chat_prompt={self.chat_prompt!r},\n"
            f"    llm={self.llm!r},\n"
            f"    parser={self.parser!r},\n"
            f"    fallback_parser={self.fallback_parser!r},\n"
            f"    chain={self.chain!r},\n"
            f"    fallback_chain={self.fallback_chain!r}\n"
            f")"
        )

    def get_chain(self) -> Runnable:
        """Retrieves the current chain.

        Returns:
            :class:`Runnable`: The current chain instance.
        """
        return self.chain

    def get_fallback_chain(self) -> Runnable:
        """Retrieve the fallback chain.

        Returns:
            :class:`Runnable`: The fallback chain instance.
        """
        return self.fallback_chain

    def _build_chain(self) -> Runnable:
        """Builds and returns a chain of `Runnable` objects.

        The chain is constructed by combining a `RunnablePassthrough` object with the `chat_prompt` and `llm` attributes. If a `parser` is provided, it is added to the chain. If `debug` is enabled, `_run_pydantic_parser_logging()` is executed.

        Returns:
            :class:`Runnable`: The constructed chain of `Runnable` objects.
        """
        # Build the base chain
        chain: Runnable = RunnablePassthrough() | self.chat_prompt | self.llm

        # Add parser if provided
        if self.parser:
            if self.debug:
                self._run_pydantic_parser_logging()
            return chain | self.parser
            
        return chain

    def _build_fallback_chain(self) -> Runnable:
        """Builds the fallback chain for the Runnable.

        This method constructs a fallback chain by combining a `RunnablePassthrough` instance with the `chat_prompt` and `llm` attributes. If a `fallback_parser` is provided, it adds the parser to the chain. If `debug` is enabled, `_run_pydantic_parser_logging()` is executed.

        Returns:
            Runnable: The constructed fallback chain.
        """
        # Build the base fallback chain
        fallback_chain: Runnable = RunnablePassthrough() | self.chat_prompt | self.llm

        # Add fallback parser if provided
        if self.fallback_parser:
            if self.debug:
                self._run_pydantic_parser_logging()
            return fallback_chain | self.fallback_parser
            
        return fallback_chain

    def _run_pydantic_parser_logging(self) -> None:
        """Performs logging related to a Pydantic parser.

        Logs the required fields and their default values (if any) for a Pydantic model if the parser is an instance of PydanticOutputParser. Additionally, it logs the JSON schema of the Pydantic model.
        """
        if isinstance(self.parser, PydanticOutputParser):
            pydantic_model: Type[BaseModel] = self.parser.pydantic_object
            self.logger.info("Required fields in Pydantic model:")
            for field_name, field in pydantic_model.model_fields.items():
                self.logger.info(
                    f"  {field_name}: {'required' if field.is_required else 'optional'}" # type: ignore[truthy-function]
                )
                if field.default is not None:
                    self.logger.info(f"    Default: {field.default}")

            self.logger.info("\nModel Schema:")
            self.logger.info(pydantic_model.model_json_schema())



from __future__ import annotations

import json
import warnings
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    cast,
    TYPE_CHECKING,
)

from langchain.schema.runnable import Runnable
from pydantic import BaseModel, ValidationError

from .._logging import get_logger, WARNING
from .._error_handling import ( # type: ignore[attr-defined]
    _ChainExceptionFactory,
    _ChainWrapperErrorReference,
)

if TYPE_CHECKING:
    from .._type_aliases import (
        ParserType,
        FallbackParserType,
    )


class _ChainWrapper:
    """A wrapper class for managing and executing chains with optional fallback chains.

    This class is designed to handle the execution of primary chains and, if necessary,
    fallback chains in case of errors or unexpected outputs. It supports preprocessing
    and postprocessing of input and output data, as well as logging for debugging and
    monitoring purposes.

    Attributes:
        chain (Runnable): The primary chain to be executed.
        fallback_chain (Runnable): The fallback chain to be executed if the primary chain fails.
        parser (ParserType | None): The parser to be used for processing the output of the primary chain.
            Defaults to None.
        fallback_parser (FallbackParserType | None): The parser to be used for processing the output of the fallback chain.
            Defaults to None.
        preprocessor (Callable | None): A function to preprocess input data before passing it to the chain.
            Defaults to None.
        postprocessor (Callable | None): A function to postprocess the output data from the chain.
            Defaults to None.
        level (logging.Level | None): The log level to use for the module.
            Defaults to logging.WARNING.
        debug (bool | None): Whether to enable debug logging.
            Defaults to False.
    """

    def __init__(
        self,
        *,
        chain: Runnable,
        fallback_chain: Runnable,
        parser: ParserType | None = None,
        fallback_parser: FallbackParserType | None = None,
        preprocessor: Callable[[Dict[str, Any]], Dict[str, Any]] | None = None,
        postprocessor: Callable[[Any], Any] | None = None,
        enable_logging: bool | None = False,
        level: int | None = WARNING,
        debug: bool | None = False,
    ) -> None:
        """Initialize the ChainBuilder.

        Args:
            chain (Runnable): The primary chain to be executed.
            fallback_chain (Runnable): The fallback chain to be executed if the primary chain fails.
            parser (ParserType | None): The parser to be used for the primary chain.
            fallback_parser (FallbackParserType | None): The parser to be used for the fallback chain.
            preprocessor (Callable | None): A function to preprocess input data before passing it to the chain.
                Defaults to None.
            postprocessor (Callable | None): A function to postprocess the output data from the chain.
                Defaults to None.
            level (logging.Level | None): The log level to use for the module.
                Defaults to logging.WARNING.
            debug (bool | None): Whether to enable debug logging.
                Defaults to False.
        """
        self.logger = get_logger(
            module_name=__name__,
            level=level,
            null_logger=not enable_logging,
        )

        self.parser: Optional[ParserType] = parser
        self.fallback_parser: Optional[FallbackParserType] = fallback_parser
        self.preprocessor: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = (
            preprocessor
        )

        self.chain: Runnable = chain
        self.fallback_chain: Runnable = fallback_chain
        self.postprocessor: Optional[Callable[[Any], Any]] = postprocessor
        if debug is None:
            warnings.warn(
                "The `debug` argument takes an optional boolean value, you gave None. "
                "The value will be set to False.\n"
                "Additionally, if your intention is to not enable debug logging, "
                "you can simply omit the `debug` argument.\n"
            )
        self.debug: bool = debug if debug is not None else False

    def __str__(self) -> str:
        """Returns a string representation of the ChainWrapper object.

        The string includes the chain, the type of the parser, and the type of the fallback parser.

        Returns:
            str: A string representation of the ChainWrapper object.
        """
        return f"ChainWrapper(chain={self.chain}, parser={type(self.parser).__name__ if self.parser else 'None'}, fallback_parser={type(self.fallback_parser).__name__ if self.fallback_parser else 'None'})"

    def __repr__(self) -> str:
        """Returns a detailed string representation for debugging.
        
        Returns:
            str: A detailed string representation of the ChainWrapper object.
        """
        return (
            f"ChainWrapper(\n"
            f"    chain={self.chain!r},\n"
            f"    parser={self.parser!r},\n"
            f"    fallback_parser={self.fallback_parser!r},\n"
            f"    preprocessor={self.preprocessor!r},\n"
            f"    postprocessor={self.postprocessor!r}\n"
            f")"
        )

    def run_chain(
        self,
        *,
        input_data: Dict[str, Any] | None = None,
        is_last_chain: bool | None = False,
    ) -> Any:
        """Executes the primary chain with the provided input data.

        If an error occurs in the primary chain, and a fallback chain is provided,
        it attempts to execute the fallback chain.

        Args:
            input_data (dict | None): The input data required to run the chain.
                Type: Dict[str, Any]
                Defaults to None.
            is_last_chain (bool | None): Indicates if this is the last chain in the sequence.
                Defaults to False.

        Returns:
            Any: The output from the chain execution, potentially processed by a postprocessor.

        Raises:
            ValueError: If no input data is provided.
            json.JSONDecodeError: If a JSON decode error occurs in both the main and fallback chains.
            ValidationError: If a validation error occurs in both the main and fallback chains.
            TypeError: If a type error occurs in both the main and fallback chains.
        """
        if is_last_chain is None:
            warnings.warn(
                "The `is_last_chain` argument takes an optional boolean value, you gave None. "
                "The value will be set to False.\n"
                "Additionally, if your intention is to not enable debug logging, "
                "you can simply omit the `debug` argument.\n"
            )
        is_last_chain: bool = ( # type: ignore[no-redef]
            False 
            if (is_last_chain is None) or (not isinstance(is_last_chain, bool))
            else is_last_chain 
        ) 
        assert isinstance(is_last_chain, bool)
        
        if (input_data is not None) and (self.preprocessor is not None):
            input_data = self.preprocessor(input_data) # type: ignore[no-redef]

        try:
            # Attempt to invoke the primary chain
            output: Any = self.chain.invoke(input_data)
        except (
            json.JSONDecodeError,
            ValidationError,
            ValueError,
            TypeError,
        ) as main_chain_exception:
            if not self._should_run_fallback_chain(is_last_chain):
                self._handle_main_chain_error(
                    main_chain_exception, 
                    is_last_chain
                )
            else:
                try:
                    # Attempt to invoke the fallback chain
                    output: Any = self.fallback_chain.invoke(input_data) # type: ignore[no-redef]
                except (
                    json.JSONDecodeError,
                    ValidationError,
                    ValueError,
                    TypeError,
                ) as fallback_exception:
                    raise _ChainExceptionFactory.create_error(
                        error=fallback_exception,
                        error_reference=_ChainWrapperErrorReference.FALLBACK_CHAIN_ERROR
                    ) from fallback_exception

        # Only need to handle intermediate chain Pydantic outputs
        # Rest are handled by JsonOutputParser or PydanticOutputParser
        if not is_last_chain and isinstance(output, BaseModel):
            return output.model_dump()

        if self.postprocessor:
            output: Any = self.postprocessor(output) # type: ignore[no-redef]

        return output

    def get_parser_type(self) -> str | None:
        """Returns the type of the parser as a string if the parser exists, otherwise returns None.

        Returns:
            str | None: The type of the parser as a string, or None if the parser
                does not exist.
        """
        return type(self.parser).__name__ if self.parser else None

    def get_fallback_parser_type(self) -> str | None:
        """Returns the type name of the fallback parser if it exists, otherwise returns None.

        Returns:
            str | None: The type name of the fallback parser as a string if it exists,
                otherwise None.
        """
        return type(self.fallback_parser).__name__ if self.fallback_parser else None
    
    def _should_run_fallback_chain(self, is_last_chain: bool) -> bool:
        """Determines if the fallback chain should be run.

        Args:
            is_last_chain (bool): Indicates if this is the last chain in the sequence.
                Defaults to False.
        """
        return (self.fallback_chain is not None) and (not is_last_chain)

    def _handle_main_chain_error(
        self,
        main_chain_exception: Exception,
        is_last_chain: bool,
    ) -> None:
        """Handles the main chain error.

        Args:
            main_chain_exception (Exception): The exception that occurred in the main chain.
            is_last_chain (bool): Indicates if this is the last chain in the sequence.
        
        Returns:
            None: if the fallback chain is provided and it is not the last chain.

        Raises:
            Exception: The exception that occurred in the main chain if there is no fallback chain or it is the last chain.
        """
        if self.fallback_chain and not is_last_chain:
            self.logger.error(
                f"Error in main chain, attempting to execute fallback chain."
            )
        elif is_last_chain:
            self.logger.error(
                f"Error in main chain and it is the last chain, raising error."
            )
            raise _ChainExceptionFactory.create_error(
                error=main_chain_exception,
                error_reference=_ChainWrapperErrorReference.MAIN_CHAIN_ERROR
            )
        else:
            self.logger.error(
                f"Error in main chain and it is not the last chain, no fallback chain provided. Raising error."
            )
            raise _ChainExceptionFactory.create_error(
                error=main_chain_exception,
                error_reference=_ChainWrapperErrorReference.MAIN_CHAIN_ERROR
            )



INPUT_ERROR_MESSAGE: str = (
    "No input data provided to ChainWrapper.run_chain().\n"
    "Input data is required to run the chain.\n"
)

MAIN_CHAIN_JSON_ERROR_MESSAGE: str = (
    "A json decode error occurred in the main chain. Consider:\n"
    "1. Check your JSON format instructions\n"
    "2. Use StrOutputParser() for intermediate chains\n"
    "3. Simplify your Pydantic model\n"
    "4. Try a more capable model\n"
)

MAIN_CHAIN_VALIDATION_ERROR_MESSAGE: str = (
    "A validation error occurred in the main chain. Consider:\n"
    "1. Check your Pydantic model\n"
    "2. Simplify your Pydantic model\n"
    "3. Try a more capable model\n"
)

MAIN_CHAIN_VALUE_ERROR_MESSAGE: str = (
    "A value error occurred in the main chain. Consider:\n"
    "1. Check your input data\n"
    "2. Simplify your Pydantic model\n"
    "3. Try a more capable model\n"
)

MAIN_CHAIN_TYPE_ERROR_MESSAGE: str = (
    "A type error occurred in the main chain. Consider:\n"
    "1. Check your input data\n"
    "2. Simplify your Pydantic model\n"
    "3. Try a more capable model\n"
)

FALLBACK_CHAIN_JSON_ERROR_MESSAGE: str = (
    "A json decode error occurred in the fallback chain. Consider:\n"
    "1. Check your JSON format instructions\n"
    "2. Use StrOutputParser() for intermediate chains\n"
    "3. Simplify your Pydantic model\n"
    "4. Try a more capable model\n"
)

FALLBACK_CHAIN_VALIDATION_ERROR_MESSAGE: str = (
    "A validation error occurred in the fallback chain. Consider:\n"
    "1. Check your Pydantic model\n"
    "2. Simplify your Pydantic model\n"
    "3. Try a more capable model\n"
)

FALLBACK_CHAIN_VALUE_ERROR_MESSAGE: str = (
    "A value error occurred in the fallback chain. Consider:\n"
    "1. Check your input data\n"
    "2. Simplify your Pydantic model\n"
    "3. Try a more capable model\n"
)

FALLBACK_CHAIN_TYPE_ERROR_MESSAGE: str = (
    "A type error occurred in the fallback chain. Consider:\n"
    "1. Check your input data\n"
    "2. Simplify your Pydantic model\n"
    "3. Try a more capable model\n"
)



from __future__ import annotations

import warnings
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    TYPE_CHECKING,
    Type,
)

from langchain.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    PromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain.schema.runnable import Runnable
from langchain_anthropic import ChatAnthropic
from langchain_core.output_parsers import (
    JsonOutputParser,
    PydanticOutputParser,
    StrOutputParser,
)
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_openai import ChatOpenAI

from .._logging import get_logger, WARNING
from .._error_handling import ( # type: ignore[attr-defined]
    _ChainExceptionFactory,
    _ChainComposerErrorReference,
    _ChainComposerValidator,
)

from .._core_internal import (
    _ChainManager,
    _ChainWrapper,
    _ChainBuilder,
)

from ..utils import APIKeyValidator

if TYPE_CHECKING:
    from pydantic import BaseModel
    from .._type_aliases import (
        FirstCallRequired,
        ParserUnion,
    )


class ChainComposer:
    """Class responsible for managing and orchestrating a sequence of chain layers,

    ChainComposer is a class responsible for managing and orchestrating a sequence of chain layers,
    each of which can process input data and produce output data. It supports various types of
    language models (LLMs) and parsers, and allows for pre-processing and post-processing of data.

    Attributes:
        api_key (str): The API key for accessing the LLM service.
        llm_model (str): The name of the LLM model to use.
        llm_model_type (str): The type of the LLM model (e.g., "openai", "anthropic", "google").
            Type: Literal["openai", "anthropic", "google"]
        llm_temperature (float): The temperature setting for the LLM model.
        llm_kwargs (dict): Additional keyword arguments for the LLM model.
            Type: Dict[str, Any]
        llm (LLMUnion): The initialized LLM instance.
        chain_manager (_ChainManager): The manager for managing the sequence of chain layers.
        chain_variables (dict): A dictionary of variables used in the chain layers.
            Type: Dict[str, Any]
        chain_variables_update_overwrite_warning_counter (int): Counter for tracking variable overwrite warnings.
        preprocessor (Callable[[Dict[str, Any]], Dict[str, Any]] | None): Optional preprocessor function.
        postprocessor (Callable[[Any], Any] | None): Optional postprocessor function.
        logger (logging.Logger): Logger for logging information and debugging.
    """

    def __init__(
        self,
        *,
        model: str,
        api_key: str,
        temperature: float = 0.7,
        preprocessor: Callable[[Dict[str, Any]], Dict[str, Any]] | None = None,
        postprocessor: Callable[[Any], Any] | None = None,
        enable_logging: bool | None = False,
        level: int | None = WARNING,
        debug: bool = False,
    ) -> None:
        """Initializes the ChainComposer class.

        Args:
            llm_model (str): The name of the language model to be used.
            api_key (str): The API key for accessing the language model.
            llm_temperature (float, optional): The temperature setting for the language model.
                Defaults to 0.7.
            preprocessor (Callable[[Dict[str, Any]], Dict[str, Any]] | None): A function to preprocess input data.
                Defaults to None.
            postprocessor (Callable[[Any], Any] | None): A function to postprocess output data.
                Defaults to None.
            llm_kwargs (dict | None): Additional keyword arguments for the language model.
                Type: Dict[str, Any]
            words_to_ban (list): A list of words to ban from the language model's output.
                Type: List[str]
                Defaults to None.
            enable_logging (bool | None): Flag to enable logging.
                Defaults to False.
            level (logging.Level | None): The logging level.
                Defaults to WARNING.
        """
        self.logger = get_logger(
            module_name=__name__,
            level=level,
            null_logger=not enable_logging,
        )

        self.api_key: str = api_key
        self.llm_model: str = model
        self.llm_model_type: str = self._get_llm_model_type(llm_model=model)

        self.llm_temperature: float = temperature
        
        # Validate API key
        self._validate_api_key(api_key=api_key)
        
        self.llm: Union[ChatOpenAI, ChatAnthropic, ChatGoogleGenerativeAI] = (
            self._initialize_llm(
                api_key=self.api_key,
                llm_model_type=self.llm_model_type,
                llm_model=self.llm_model,
                llm_temperature=self.llm_temperature,
            )
        )
        self.enable_logging: bool = ( # type: ignore[assignment]
            False
            if (
                (enable_logging is False) or 
                (enable_logging is None) or 
                (not isinstance(enable_logging, bool))
            )
            else enable_logging
        )
        self.level: int = (  # type: ignore[assignment]
            WARNING
            if (
                (level is None) or 
                (not isinstance(level, int))
            )
            else level
        )
        self.chain_manager: _ChainManager = _ChainManager(
            enable_logging=self.enable_logging,
            level=self.level,
        )

        self.chain_variables: Dict[str, Any] = {}

        self.chain_variables_update_overwrite_warning_counter: int = 0
        self.preprocessor: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = (
            preprocessor
        )
        self.postprocessor: Optional[Callable[[Any], Any]] = postprocessor
        self.debug: bool = debug

    def __str__(self) -> str:
        """Returns a concise string representation of the ChainComposer object.

        Returns:
            str: A concise string representation of the ChainComposer object.
        """
        return (
            f"ChainComposer(\n"
            f"  llm={self.llm.__class__.__name__},\n"
            f"  model={self.llm_model},\n"
            f"  temperature={self.llm_temperature},\n"
            f"  chain_count={len(self.chain_manager.chain_sequence)}\n"
            f")"
        )

    def __repr__(self) -> str:
        """Returns a detailed string representation for debugging.

        Returns:
            str: A detailed string representation of the ChainComposer object.
        """
        return (
            f"ChainComposer(\n"
            f"    llm={self.llm!r},\n"
            f"    llm_model={self.llm_model!r},\n"
            f"    llm_model_type={self.llm_model_type!r},\n"
            f"    llm_temperature={self.llm_temperature!r},\n"
            f"    chain_manager={self.chain_manager!r},\n"
            f"    chain_variables={self.chain_variables!r},\n"
            f"    preprocessor={self.preprocessor!r},\n"
            f"    postprocessor={self.postprocessor!r}\n"
            f")"
        )
    
    def add_chain_layer(
        self,
        *,
        system_prompt: str,
        human_prompt: str,
        output_passthrough_key_name: str | None = None,
        ignore_output_passthrough_key_name_error: bool = False,
        preprocessor: Callable[[Dict[str, Any]], Dict[str, Any]] | None = None,
        postprocessor: Callable[[Any], Any] | None = None,
        parser_type: Literal["pydantic", "json", "str"] | None = None,
        fallback_parser_type: Literal["pydantic", "json", "str"] | None = None,
        pydantic_output_model: Type[BaseModel] | None = None,
        fallback_pydantic_output_model: Type[BaseModel] | None = None,
    ) -> "ChainComposer":
        """Adds a chain layer to the chain composer.

        This method configures and adds a new chain layer to the chain composer,
        allowing for the processing of input data through specified prompts and parsers.

        Args:
            system_prompt (str): The system prompt template for the chain layer.
            human_prompt (str): The human prompt template for the chain layer.
            output_passthrough_key_name (str | None): Key name for passing chain output
                to the next layer.
                Defaults to None.
            ignore_output_passthrough_key_name_error (bool): Flag to ignore missing output
                key name errors.
                Defaults to False.
            preprocessor (Callable[[Dict[str, Any]], Dict[str, Any]] | None): Function to preprocess input data.
                Defaults to None.
            postprocessor (Callable[[Any], Any] | None): Function to postprocess output data.
                Defaults to None.
            parser_type (str | None): Type of parser to use.
                Type: Literal["pydantic", "json", "str"] | None
                Defaults to None.
            fallback_parser_type (str | None): Type of fallback parser.
                Type: Literal["pydantic", "json", "str"] | None
                Defaults to None.
            pydantic_output_model (Type[BaseModel] | None): Pydantic model for output validation.
                Defaults to None
            fallback_pydantic_output_model (Type[BaseModel] | None): Pydantic model for fallback parser.
                Defaults to None.

        Returns:
            None
        """
        self._run_chain_validation_checks(
            output_passthrough_key_name=output_passthrough_key_name,
            ignore_output_passthrough_key_name_error=ignore_output_passthrough_key_name_error,
            parser_type=parser_type,
            pydantic_output_model=pydantic_output_model,
            fallback_parser_type=fallback_parser_type,
            fallback_pydantic_output_model=fallback_pydantic_output_model,
        )

        parser: Optional[ParserUnion] = None
        fallback_parser: Optional[ParserUnion] = None
        if parser_type:
            parser = self._initialize_parser(
                parser_type=parser_type, pydantic_output_model=pydantic_output_model
            )
        if fallback_parser_type:
            fallback_parser = self._initialize_parser(
                parser_type=fallback_parser_type,
                pydantic_output_model=fallback_pydantic_output_model,
            )
        # Create prompt templates without specifying input_variables
        system_prompt_template: PromptTemplate = PromptTemplate(template=system_prompt) # type: ignore[call-arg]
        human_prompt_template: PromptTemplate = PromptTemplate(template=human_prompt) # type: ignore[call-arg]
        system_message_prompt_template: SystemMessagePromptTemplate = (
            SystemMessagePromptTemplate.from_template(system_prompt_template.template)
        )
        human_message_prompt_template: HumanMessagePromptTemplate = (
            HumanMessagePromptTemplate.from_template(human_prompt_template.template)
        )

        chat_prompt_template: ChatPromptTemplate = ChatPromptTemplate.from_messages(
            [system_message_prompt_template, human_message_prompt_template]
        )
        # Build the chain using ChainBuilder
        chain_builder: _ChainBuilder = _ChainBuilder(
            chat_prompt=chat_prompt_template,
            llm=self.llm,
            parser=parser,
            fallback_parser=fallback_parser,
            enable_logging=self.enable_logging,
            level=self.level,
        )
        chain: Runnable = chain_builder.get_chain()
        fallback_chain: Runnable = chain_builder.get_fallback_chain()

        # Wrap the chain
        chain_wrapper: _ChainWrapper = _ChainWrapper(
            chain=chain,
            fallback_chain=fallback_chain,
            parser=parser,
            fallback_parser=fallback_parser,
            preprocessor=preprocessor or self.preprocessor,
            postprocessor=postprocessor or self.postprocessor,
            enable_logging=self.enable_logging,
            level=self.level,
        )

        # Add the chain to the composer
        self.chain_manager.add_chain(
            chain_wrapper=chain_wrapper,
            output_passthrough_key_name=output_passthrough_key_name,
        )

        # Return the ChainManager instance to allow for method chaining
        return self
    
    def run(
        self,
        prompt_variables_dict: Union[FirstCallRequired, None] = None,
    ) -> Dict[str, Any]:
        """Executes the chain builder process.

        This method performs validation checks, updates chain variables if provided,
        and runs the chain composer with the current chain variables.

        Args:
            prompt_variables_dict (dict | None): A dictionary containing prompt variables.
                If provided, it will be used to update the chain variables.
                Type: Union[FirstCallRequired, None]
                Defaults to None.

        Returns:
            dict: The result of running the chain composer.
        """
        self._run_validation_checks(
            prompt_variables_dict=prompt_variables_dict,
        )

        if prompt_variables_dict is not None:
            self._update_chain_variables(prompt_variables_dict)

        return self.chain_manager.run(
            data_dict=self.chain_variables,
            data_dict_update_function=self._update_chain_variables,
        )
        
    def get_chain_sequence(self) -> List[Tuple[_ChainWrapper, str | None]]:
        """Retrieves the chain sequence from the chain composer.

        Returns:
            list: A list of tuples where each tuple contains a ChainWrapper object and
                the output key name if output_passthrough_key_name was provided to add_chain_layer.
                Type: List[Tuple[ChainWrapper, str | None]]
        """
        return self.chain_manager.chain_sequence

    def print_chain_sequence(self) -> None:
        """Prints the chain sequence by formatting it.

        This method retrieves the chain sequence from the chain composer and
        formats it using the _format_chain_sequence method.

        Returns:
            None
        """
        chain_sequence: List[Tuple[_ChainWrapper, str | None]] = (
            self.chain_manager.chain_sequence
        )
        self._format_chain_sequence(chain_sequence)

    def get_chain_variables(self) -> Dict[str, Any]:
        """Retrieve the chain variables.

        Returns:
            dict: A dictionary containing the chain variables.
                Type: Dict[str, Any]
        """
        return self.chain_variables

    def print_chain_variables(self) -> None:
        """Prints the chain variables in a formatted manner.

        This method prints the chain variables stored in the `chain_variables`
        attribute of the class. The output is formatted with a header and
        footer consisting of dashes, and each key-value pair is printed on
        a new line.

        Returns:
            None
        """
        print(f"Chain Variables:\n{'-' * 10}")
        for key, value in self.chain_variables.items():
            print(f"{key}: {value}")
        print(f"{'-' * 10}\n")

    def _get_llm_model_type(self, *, llm_model: str) -> str:
        """Determine the type of LLM (Large Language Model) based on the provided model name.

        Args:
            llm_model (str): The name of the LLM model.

        Returns:
            str: The type of the LLM model. Possible values are "openai", "anthropic", and "google".

        Raises:
            ChainError: If the provided LLM model name does not match any of the supported types.
        """
        return _ChainComposerValidator.validate_llm_model_type(llm_model)

    def _initialize_llm(
        self,
        *,
        api_key: str | None = None,
        llm_model_type: str | None = None,
        llm_model: str | None = None,
        llm_temperature: float | None = None,
    ) -> Union[ChatOpenAI, ChatAnthropic, ChatGoogleGenerativeAI]:
        """Initializes a language model based on the specified type.

        Args:
            api_key (str | None): The API key for accessing the language model service.
                Defaults to None.
            llm_model_type (str | None): The type of the language model (e.g., "openai", "anthropic", "google").
                Defaults to None.
            llm_model (str | None): The specific model to use within the chosen type.
                Defaults to None.
            llm_temperature (float | None): The temperature setting for the language model, affecting randomness.
                Defaults to None.

        Returns:
            LLMUnion: An instance of the initialized language model.

        Raises:
            ChainError: If the specified `llm_model_type` is not supported.
        """
        if api_key is None:
            api_key = self.api_key

        if llm_model_type is None:
            llm_model_type = self.llm_model_type

        if llm_model is None:
            llm_model = self.llm_model

        if llm_temperature is None:
            llm_temperature = self.llm_temperature

        if llm_model_type == "openai":
            return self._create_openai_llm(
                api_key, llm_model, llm_temperature
            )
        elif llm_model_type == "anthropic":
            return self._create_anthropic_llm(
                api_key, llm_model, llm_temperature
            )
        elif llm_model_type == "google":
            return self._create_google_llm(
                api_key, llm_model, llm_temperature
            )
        else:
            raise _ChainExceptionFactory.create_error(
                error=ValueError(
                    f"Unsupported LLM model type: {llm_model_type}. Supported types: openai, anthropic, google."
                ),
                error_reference=_ChainComposerErrorReference.UNSUPPORTED_LLM_MODEL_TYPE
            )

    def _recreate_llm(self) -> None:
        """Recreates the LLM with the current parameters."""
        self.llm = self._initialize_llm(
            api_key=self.api_key,
            llm_model_type=self.llm_model_type,
            llm_model=self.llm_model,
            llm_temperature=self.llm_temperature,
        )

    def _create_openai_llm(
        self,
        api_key: str,
        llm_model: str,
        llm_temperature: float,
    ) -> ChatOpenAI:
        """Creates an instance of the ChatOpenAI language model.

        Args:
            api_key (str): The API key for authenticating with the OpenAI service.
            llm_model (str): The identifier of the language model to use.
            llm_temperature (float): The temperature setting for the language model,
                which controls the randomness of the output.

        Returns:
            ChatOpenAI: An instance of the ChatOpenAI language model configured with
                the specified parameters.
        """
        return ChatOpenAI(
            model=llm_model,
            api_key=api_key, # type: ignore[arg-type]
            temperature=llm_temperature,
        )

    def _create_anthropic_llm(
        self, 
        api_key: str, 
        llm_model: str, 
        llm_temperature: float,
    ) -> ChatAnthropic:
        """Creates an instance of the ChatAnthropic language model.

        Args:
            api_key (str): The API key for authenticating with the Anthropic service.
            llm_model (str): The identifier of the language model to use.
            llm_temperature (float): The temperature setting for the language model,
                controlling the randomness of the output.

        Returns:
            ChatAnthropic: An instance of the ChatAnthropic language model.
        """
        return ChatAnthropic( # type: ignore[call-arg]
            model_name=llm_model, 
            api_key=api_key, # type: ignore[arg-type]
            temperature=llm_temperature,
        )

    def _create_google_llm(
        self, 
        api_key: str, 
        llm_model: str, 
        llm_temperature: float,
    ) -> ChatGoogleGenerativeAI:
        """Creates an instance of ChatGoogleGenerativeAI with the specified parameters.

        Args:
            api_key (str): The API key for authenticating with the Google LLM service.
            llm_model (str): The model identifier for the Google LLM.
            llm_temperature (float): The temperature setting for the LLM,
                which controls the randomness of the output.

        Returns:
            ChatGoogleGenerativeAI: An instance of the ChatGoogleGenerativeAI class configured
                with the provided parameters.
        """
        return ChatGoogleGenerativeAI(
            model=llm_model, 
            api_key=api_key, # type: ignore[arg-type]
            temperature=llm_temperature,
        )

    def _initialize_parser(
        self,
        parser_type: Literal["pydantic", "json", "str"],
        pydantic_output_model: Type[BaseModel] | None = None,
    ) -> ParserUnion:
        """Initializes and returns a parser based on the specified parser type.

        Args:
            parser_type (str): The type of parser to initialize.
                Must be one of "pydantic", "json", or "str".
                Type: Literal["pydantic", "json", "str"]
            pydantic_output_model (BaseModel | None): The Pydantic model to use for the parser. Required if parser_type is "pydantic".
                Defaults to None.

        Returns:
            ParserUnion: An instance of the specified parser type.

        Raises:
            ChainError: If an invalid parser_type is provided.
        """
        if parser_type == "pydantic":
            parser: PydanticOutputParser = self._create_pydantic_parser( # type: ignore[no-redef]
                pydantic_output_model=pydantic_output_model
            )
            self.logger.debug(f"Created Pydantic parser: {parser}")
            return parser
        elif parser_type == "json":
            parser: JsonOutputParser = self._create_json_parser( # type: ignore[no-redef]
                pydantic_output_model=pydantic_output_model
            )
            self.logger.debug(f"Created JSON parser: {parser}")
            return parser
        elif parser_type == "str":
            parser: StrOutputParser = self._create_str_parser() # type: ignore[no-redef]
            self.logger.debug(f"Created Str parser: {parser}")
            return parser
        else:
            raise _ChainExceptionFactory.create_error(
                error=ValueError(f"Invalid parser_type: {parser_type}"),
                error_reference=_ChainComposerErrorReference.INVALID_PARSER_TYPE
            )

    def _create_pydantic_parser(
        self,
        pydantic_output_model: Type[BaseModel] | None,
    ) -> PydanticOutputParser:
        """Creates a Pydantic output parser.

        Args:
            pydantic_output_model (BaseModel | None): The Pydantic model to be used for parsing output.

        Returns:
            PydanticOutputParser: An instance of PydanticOutputParser initialized with
                the provided Pydantic model.

        Raises:
            ChainError: If pydantic_output_model is not provided.

        Notes:
            - `pydantic_output_model` must be provided for `parser_type` 'pydantic'.
        """
        if not pydantic_output_model:
            raise _ChainExceptionFactory.create_error(
                error=ValueError(
                    "pydantic_output_model must be provided for 'pydantic' parser_type."
                ),
                error_reference=_ChainComposerErrorReference.MISSING_PYDANTIC_MODEL
            )
        return PydanticOutputParser(pydantic_object=pydantic_output_model)

    def _create_json_parser(
        self, *, pydantic_output_model: Type[BaseModel] | None
    ) -> JsonOutputParser:
        """Creates a JSON parser for the chain layer output.

        Args:
            pydantic_output_model (Type[BaseModel] | None): An optional Pydantic model to enforce
                typing on the JSON output.
                Type: Type[BaseModel] | None
                Defaults to None.

        Returns:
            JsonOutputParser: An instance of JsonOutputParser. If pydantic_output_model is provided, the parser will enforce the model's schema on the output.

        Raises:
            UserWarning: If `pydantic_output_model` is not provided, a warning is issued
                recommending its use for proper typing of the output.
        """
        json_parser: JsonOutputParser | None = None
        if not pydantic_output_model:
            warnings.warn(
                "It is highly recommended to provide a pydantic_output_model when parser_type is 'json'. "
                "This will ensure that the output of the chain layer is properly typed and can be used in downstream chain layers."
            )
            self.logger.debug("Creating JSON parser without pydantic_output_model. ")
            json_parser = JsonOutputParser()
        else:
            self.logger.debug(
                f"Creating JSON parser with pydantic_output_model: {pydantic_output_model}"
            )
        return JsonOutputParser(pydantic_object=pydantic_output_model)

    def _create_str_parser(self) -> StrOutputParser:
        """Creates an instance of StrOutputParser.

        Returns:
            StrOutputParser: An instance of the StrOutputParser class.
        """
        return StrOutputParser()

    def _run_chain_validation_checks(
        self,
        *,
        output_passthrough_key_name: str | None,
        ignore_output_passthrough_key_name_error: bool,
        parser_type: Literal["pydantic", "json", "str"] | None,
        pydantic_output_model: Type[BaseModel] | None,
        fallback_parser_type: Literal["pydantic", "json", "str"] | None,
        fallback_pydantic_output_model: Type[BaseModel] | None,
    ) -> None:
        """Validates chain configuration parameters before execution.

        Performs validation checks on chain configuration parameters to ensure proper setup
        and compatibility between different components.

        Args:
            output_passthrough_key_name (str | None): Optional key name for passing chain
                output to next layer.
                Defaults to None.
            ignore_output_passthrough_key_name_error (bool): Whether to ignore missing output
                key name errors.
                Defaults to False.
            parser_type (str | None): Type of parser to use.
                Type: Literal["pydantic", "json", "str"] | None
                Defaults to None.
            pydantic_output_model (Type[BaseModel] | None): Pydantic model for output validation.
                Defaults to None.
            fallback_parser_type (str | None): Type of fallback parser.
                Type: Literal["pydantic", "json", "str"] | None
                Defaults to None.
            fallback_pydantic_output_model (Type[BaseModel] | None): Pydantic model for fallback parser.
                Defaults to None.

        Raises:
            ChainError: If validation fails for:
                - Missing output key name when required
                - Invalid parser type combinations
                - Missing required models
                - Duplicate parser types
                - Same models used for main and fallback

        Warnings:
            UserWarning: For non-critical issues like:
                - Missing output key name when ignored
                - Missing recommended Pydantic models
                - Unused provided models
        """
        _ChainComposerValidator.validate_parser_configuration(
            output_passthrough_key_name=output_passthrough_key_name,
            ignore_output_passthrough_key_name_error=ignore_output_passthrough_key_name_error,
            parser_type=parser_type,
            pydantic_output_model=pydantic_output_model,
            fallback_parser_type=fallback_parser_type,
            fallback_pydantic_output_model=fallback_pydantic_output_model,
        )

    def _format_chain_sequence(
        self, chain_sequence: List[Tuple[_ChainWrapper, str | None]]
    ) -> None:
        """Formats and prints the details of each chain in the given chain sequence.

        Args:
            chain_sequence (list): A list of tuples where each tuple contains a ChainWrapper
                object and an optional output name.
                Type: List[Tuple[ChainWrapper, str | None]]

        Returns:
            None
        """
        for index, (chain_wrapper, output_name) in enumerate(chain_sequence):
            print(f"Chain {index + 1}:")
            print(f"\tOutput Name: {output_name}")
            print(f"\tParser Type: {chain_wrapper.get_parser_type()}")
            print(f"\tFallback Parser Type: {chain_wrapper.get_fallback_parser_type()}")
            print(f"\tPreprocessor: {chain_wrapper.preprocessor}")
            print(f"\tPostprocessor: {chain_wrapper.postprocessor}")

    def _run_validation_checks(
        self,
        *,
        prompt_variables_dict: Union[FirstCallRequired, None],
    ) -> None:
        """Validates the input parameters for the chain execution.

        Args:
            prompt_variables_dict (dict | None): A dictionary containing the variables
                to be passed to the chain layers.
                Type: Union[FirstCallRequired, None]
                - On the first call to `run()`, this parameter must be provided.
                - On subsequent calls, it can be omitted if there are no new variables to pass.

        Raises:
            ChainError: If `prompt_variables_dict` is None on the first call to `run()`.
            TypeError: If `prompt_variables_dict` is not a dictionary when provided.

        Notes:
            - The `prompt_variables_dict` should contain keys that match the variable names
              used in the chain layers.
            - The `output_passthrough_key_name` parameter in the `add_chain_layer` method is
              used to identify the output of the chain layer and assign it to a variable.
            - If `output_passthrough_key_name` is not specified, the output of the chain layer
              will not be assigned to a variable and will not be available to the next chain layer.
            - The `ignore_output_passthrough_key_name_error` parameter can be set to True if
              the output of the chain layer is not needed for the next chain layer, such as
              when running a chain layer solely for its side effects or if it is the last
              chain layer in a multi-layer chain.
            - Ensure that the placeholder variable names in your prompt strings match the keys
              in `prompt_variables_dict` passed into the `ChainManager.run()` method.
        """
        _ChainComposerValidator.validate_prompt_variables(
            prompt_variables_dict=prompt_variables_dict,
            is_first_call=self.chain_variables_update_overwrite_warning_counter == 0,
        )

    def _format_overwrite_warning(self, overwrites: Dict[str, Dict[str, Any]]) -> str:
        """Formats a warning message for overwritten values.

        Args:
            overwrites (dict): A dictionary where the key is the name of the overwritten item,
                and the value is another dictionary with 'old' and 'new' keys representing
                the old and new values respectively.
                Type: Dict[str, Dict[str, Any]]

        Returns:
            str: A formatted string that lists each overwritten item with its old and
                new values.
        """
        return "\n".join(
            f"  {key}:\n    - {details['old']}\n    + {details['new']}"
            for key, details in overwrites.items()
        )

    def _check_first_time_overwrites(
        self, prompt_variables_dict: Dict[str, Any]
    ) -> None:
        """Checks and warns if any global chain variables are being overwritten for the first time.

        This method compares the keys in the provided `prompt_variables_dict` with the existing
        `chain_variables`. If any keys match, it indicates that an overwrite is occurring. A warning
        is issued the first time this happens, detailing the old and new values of the overwritten
        variables. Subsequent overwrites will not trigger warnings.

        Args:
            prompt_variables_dict (dict): A dictionary containing the new values for the chain
                variables that may overwrite existing ones.
                Type: Dict[str, Any]

        Returns:
            None
        """
        if self.chain_variables_update_overwrite_warning_counter == 0:
            # overwrites = {
            #     key: {
            #         "old": self.chain_variables[key],
            #         "new": prompt_variables_dict[key],
            #     }
            #     for key in prompt_variables_dict
            #     if key in self.chain_variables
            # }
            # if overwrites:
            #     warnings.warn(
            #         f"Overwriting existing global variables:\n"
            #         f"{self._format_overwrite_warning(overwrites)}\n"
            #         "Subsequent overwrites will not trigger warnings."
            #     )
            self.chain_variables_update_overwrite_warning_counter += 1

    def _update_chain_variables(self, variables: Dict[str, Any]) -> None:
        """Update chain variables.

        Args:
            variables (Dict[str, Any]): The variables to update.
        """
        for key, value in variables.items():
            if key in self.chain_variables:
                if self.debug:
                    warnings.warn(
                        f"Overwriting existing chain variable '{key}' with new value '{value}'",
                        UserWarning,
                        stacklevel=2
                    )
            self.chain_variables[key] = value

    def _validate_api_key(self, api_key: str) -> None:
        """Validate the API key.

        Args:
            api_key (str): The API key to validate.

        Raises:
            ChainError: If the API key is invalid.
        """
        if not api_key:
            raise _ChainExceptionFactory.create_error(
                error=ValueError("API key cannot be empty"),
                error_reference=_ChainComposerErrorReference.INVALID_API_KEY
            )

        validator = APIKeyValidator()
        if not validator.is_valid(api_key=api_key, model=self.llm_model):
            raise _ChainExceptionFactory.create_error(
                error=ValueError("Invalid API key"),
                error_reference=_ChainComposerErrorReference.INVALID_API_KEY
            )




from .log_helpers import (
    get_logger,
    DEBUG,
    INFO,
    WARNING,
    ERROR,
    CRITICAL,
)

__all__ = [
    "get_logger",
    "DEBUG",
    "INFO",
    "WARNING",
    "ERROR",
    "CRITICAL",
]



from __future__ import annotations

from typing import Union, TYPE_CHECKING, cast, Type

from ..enums._error_reference import _ChainComposerErrorReference
from ..factories import _ChainExceptionFactory

if TYPE_CHECKING:
    from pydantic import BaseModel
    from chain_composer._type_aliases import FirstCallRequired

class _ChainComposerValidator:
    """Validator class for ChainComposer operations."""
    
    @staticmethod
    def validate_llm_model_type(llm_model: str) -> str:
        """Validate and determine the LLM model type.
        
        Args:
            llm_model (str): The name of the LLM model.
            
        Returns:
            str: The validated model type.
            
        Raises:
            ChainError: If the model type is not supported.
        """
        if llm_model.lower().startswith("gpt"):
            return "openai"
        elif llm_model.lower().startswith("claude"):
            return "anthropic"
        elif llm_model.lower().startswith("gemini"):
            return "google"
        
        raise _ChainExceptionFactory.create_error(
            error=ValueError(f"Unsupported LLM model: {llm_model}"),
            error_reference=_ChainComposerErrorReference.UNSUPPORTED_LLM_MODEL
        )
    
    @staticmethod
    def validate_parser_configuration(
        *,
        output_passthrough_key_name: str | None,
        ignore_output_passthrough_key_name_error: bool,
        parser_type: str | None,
        pydantic_output_model: Type[BaseModel] | None,
        fallback_parser_type: str | None,
        fallback_pydantic_output_model: Type[BaseModel] | None,
    ) -> None:
        """Validate parser configuration parameters.
        
        Args:
            output_passthrough_key_name (str | None): Key name for output.
            ignore_output_passthrough_key_name_error (bool): Whether to ignore missing key name.
            parser_type (str | None): Type of parser.
            pydantic_output_model (Type[BaseModel] | None): Pydantic model for output.
            fallback_parser_type (str | None): Type of fallback parser.
            fallback_pydantic_output_model (Type[BaseModel] | None): Pydantic model for fallback.
            
        Raises:
            ChainError: If validation fails.
        """
        if not output_passthrough_key_name and not ignore_output_passthrough_key_name_error:
            raise _ChainExceptionFactory.create_error(
                error=ValueError("Missing output key name"),
                error_reference=_ChainComposerErrorReference.MISSING_OUTPUT_KEY
            )
            
        if parser_type is None and fallback_parser_type is not None:
            raise _ChainExceptionFactory.create_error(
                error=ValueError("Invalid parser combination"),
                error_reference=_ChainComposerErrorReference.INVALID_PARSER_COMBINATION
            )
            
        if parser_type == "pydantic" and not pydantic_output_model:
            raise _ChainExceptionFactory.create_error(
                error=ValueError("Missing Pydantic model for parser"),
                error_reference=_ChainComposerErrorReference.MISSING_PYDANTIC_MODEL
            )
            
        if fallback_parser_type == "pydantic" and not fallback_pydantic_output_model:
            raise _ChainExceptionFactory.create_error(
                error=ValueError("Missing Pydantic model for fallback parser"),
                error_reference=_ChainComposerErrorReference.MISSING_PYDANTIC_MODEL
            )
            
        if parser_type and parser_type not in ["pydantic", "json", "str"]:
            raise _ChainExceptionFactory.create_error(
                error=ValueError(f"Invalid parser type: {parser_type}"),
                error_reference=_ChainComposerErrorReference.INVALID_PARSER_TYPE
            )
            
        if fallback_parser_type and fallback_parser_type not in ["pydantic", "json", "str"]:
            raise _ChainExceptionFactory.create_error(
                error=ValueError(f"Invalid fallback parser type: {fallback_parser_type}"),
                error_reference=_ChainComposerErrorReference.INVALID_PARSER_TYPE
            )
    
    @staticmethod
    def validate_prompt_variables(
        prompt_variables_dict: Union[FirstCallRequired, None],
        is_first_call: bool,
    ) -> None:
        """Validate prompt variables.
        
        Args:
            prompt_variables_dict (Union[FirstCallRequired, None]): The variables dict.
            is_first_call (bool): Whether this is the first call to run().
            
        Raises:
            ChainError: If validation fails.
        """
        if is_first_call and prompt_variables_dict is None:
            raise _ChainExceptionFactory.create_error(
                error=ValueError("First call must provide prompt variables"),
                error_reference=_ChainComposerErrorReference.MISSING_PROMPT_VARIABLES
            )
            
        if prompt_variables_dict is not None:
            if not isinstance(cast(dict, prompt_variables_dict), dict):
                raise _ChainExceptionFactory.create_error(
                    error=TypeError(
                        "prompt_variables_dict must be a dictionary. "
                        "Each key should match the variable names used in your chain layers. "
                        "output_passthrough_key_name parameter in add_chain_layer method is used to identify the output "
                        "of the chain layer "
                        "and assign it to a variable. If you do not specify output_passthrough_key_name, the output of "
                        "the chain layer will not be assigned to a variable and thus will not be available to the next "
                        "chain layer. "
                        "If you do not need the output of the chain layer to be passed to the next chain layer, you can "
                        "set ignore_output_passthrough_key_name_error to True. "
                        "A time to set ignore_output_passthrough_key_name_error to True is when you are running a chain "
                        "layer solely for its side effects (e.g. printing, saving to a database, etc.) without needing the "
                        "output of the chain layer to be passed to the next chain layer. "
                        "Another reason to set ignore_output_passthrough_key_name_error to True is if you have a "
                        "multi-layer chain and this is your last chain layer. "
                        "Check your prompt strings for your placeholder variables, these names should match the keys in "
                        f"{prompt_variables_dict} passed into the ChainManager.run() method."
                    ),
                    error_reference=_ChainComposerErrorReference.INVALID_PROMPT_VARIABLES
                ) 



from ._llm_validation_error import _LLMValidationError
from ._llm_json_decode_error import _LLMJsonDecodeError
from ._base_chain_error import _ChainError

__all__ = [
    "_ChainError",
    "_LLMJsonDecodeError",
    "_LLMValidationError",
]


from enum import Enum

class _ErrorType(Enum):
    JSON_ERROR = "json_error"
    VALIDATION_ERROR = "validation_error"
    VALUE_ERROR = "value_error"
    TYPE_ERROR = "type_error"
    UNKNOWN_ERROR = "unknown_error"


from typing import (
    Any,
    Dict,
    TypeVar,
    Union,
    TypeAlias,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from langchain_core.output_parsers import (
        JsonOutputParser,
        PydanticOutputParser,
        StrOutputParser,
    )
    
    from langchain_google_genai import ChatGoogleGenerativeAI
    from langchain_openai import ChatOpenAI
    from langchain_anthropic import ChatAnthropic

FirstCallRequired = TypeVar("FirstCallRequired", bound=Dict[str, Any])
"""TypeVar representing a dictionary that is required on first call but optional after.

Type Structure:
    TypeVar bound to Dict[str, Any]

Usage:
    Used to enforce that a dictionary parameter must be provided on first call
    but can be optional on subsequent calls.
"""

ParserUnion: TypeAlias = Union[PydanticOutputParser, JsonOutputParser, StrOutputParser]
"""Union type representing valid parser types.

Type Structure:
    Union[PydanticOutputParser, JsonOutputParser, StrOutputParser]
"""

ParserType = TypeVar("ParserType", bound=ParserUnion)
"""TypeVar representing the main parser type.

Type Structure:
    TypeVar bound to Union[PydanticOutputParser, JsonOutputParser, StrOutputParser]

Usage:
    Used to enforce type consistency for the primary parser in chain operations.
"""

FallbackParserType = TypeVar("FallbackParserType", bound=ParserUnion)
"""TypeVar representing the fallback parser type.

Type Structure:
    TypeVar bound to Union[PydanticOutputParser, JsonOutputParser, StrOutputParser]

Usage:
    Used to enforce type consistency for the fallback parser in chain operations.
"""

LLMUnion = TypeVar("LLMUnion", bound=Union[ChatOpenAI, ChatAnthropic, ChatGoogleGenerativeAI])
"""TypeVar representing the language model type.

Type Structure:
    TypeVar bound to Union[ChatOpenAI, ChatAnthropic, ChatGoogleGenerativeAI]

Usage:
    Used to enforce type consistency for the language model in chain operations.
"""



import json
import pytest
from pydantic import ValidationError

from chain_composer._error_handling import (
    _ChainError,
    _ErrorType,
    _ChainExceptionFactory,
    _ChainComposerValidator,
    _ChainComposerErrorReference,
    _ChainWrapperErrorReference,
)

def test_error_type_enum():
    """Test ErrorType enum values"""
    assert _ErrorType.JSON_ERROR.value == "json_error"
    assert _ErrorType.VALIDATION_ERROR.value == "validation_error"
    assert _ErrorType.VALUE_ERROR.value == "value_error"
    assert _ErrorType.TYPE_ERROR.value == "type_error"
    assert _ErrorType.UNKNOWN_ERROR.value == "unknown_error"

def test_chain_composer_error_reference():
    """Test ChainComposerErrorReference enum values"""
    assert _ChainComposerErrorReference.UNSUPPORTED_LLM_MODEL.value == "unsupported_llm_model"
    assert _ChainComposerErrorReference.INVALID_PARSER_TYPE.value == "invalid_parser_type"
    assert _ChainComposerErrorReference.MISSING_PYDANTIC_MODEL.value == "missing_pydantic_model"
    assert _ChainComposerErrorReference.MISSING_OUTPUT_KEY.value == "missing_output_key"
    assert _ChainComposerErrorReference.INVALID_PARSER_COMBINATION.value == "invalid_parser_combination"
    assert _ChainComposerErrorReference.MISSING_PROMPT_VARIABLES.value == "missing_prompt_variables"
    assert _ChainComposerErrorReference.INVALID_PROMPT_VARIABLES.value == "invalid_prompt_variables"
    assert _ChainComposerErrorReference.INVALID_API_KEY.value == "invalid_api_key"

def test_chain_wrapper_error_reference():
    """Test ChainWrapperErrorReference enum values"""
    assert _ChainWrapperErrorReference.INPUT_ERROR.value == "input_error"
    assert _ChainWrapperErrorReference.MAIN_CHAIN_ERROR.value == "main_chain_error"
    assert _ChainWrapperErrorReference.FALLBACK_CHAIN_ERROR.value == "fallback_chain_error"

def test_chain_exception_factory_json_error():
    """Test ChainExceptionFactory with JSON error"""
    error = json.JSONDecodeError("test", "test", 0)
    chain_error = _ChainExceptionFactory.create_error(
        error=error,
        error_reference=_ChainWrapperErrorReference.MAIN_CHAIN_ERROR
    )
    assert isinstance(chain_error, _ChainError)

def test_chain_exception_factory_validation_error():
    """Test ChainExceptionFactory with validation error"""
    error = ValidationError.from_exception_data("test", [])
    chain_error = _ChainExceptionFactory.create_error(
        error=error,
        error_reference=_ChainWrapperErrorReference.MAIN_CHAIN_ERROR
    )
    assert isinstance(chain_error, _ChainError)

def test_chain_exception_factory_value_error():
    """Test ChainExceptionFactory with value error"""
    error = ValueError("test error")
    chain_error = _ChainExceptionFactory.create_error(
        error=error,
        error_reference=_ChainWrapperErrorReference.MAIN_CHAIN_ERROR
    )
    assert isinstance(chain_error, _ChainError)

def test_chain_exception_factory_type_error():
    """Test ChainExceptionFactory with type error"""
    error = TypeError("test error")
    chain_error = _ChainExceptionFactory.create_error(
        error=error,
        error_reference=_ChainWrapperErrorReference.MAIN_CHAIN_ERROR
    )
    assert isinstance(chain_error, _ChainError)

def test_chain_exception_factory_unknown_error():
    """Test ChainExceptionFactory with unknown error"""
    error = Exception("test error")
    chain_error = _ChainExceptionFactory.create_error(
        error=error,
        error_reference=_ChainWrapperErrorReference.MAIN_CHAIN_ERROR
    )
    assert isinstance(chain_error, _ChainError)

def test_chain_composer_validator_llm_model_type():
    """Test ChainComposerValidator LLM model type validation"""
    # Test valid models
    assert _ChainComposerValidator.validate_llm_model_type("gpt-4") == "openai"
    assert _ChainComposerValidator.validate_llm_model_type("claude-2") == "anthropic"
    assert _ChainComposerValidator.validate_llm_model_type("gemini-pro") == "google"
    
    # Test invalid model
    with pytest.raises(_ChainError):
        _ChainComposerValidator.validate_llm_model_type("invalid_model")

def test_chain_composer_validator_parser_configuration():
    """Test ChainComposerValidator parser configuration validation"""
    # Test missing output key
    with pytest.raises(_ChainError):
        _ChainComposerValidator.validate_parser_configuration(
            output_passthrough_key_name=None,
            ignore_output_passthrough_key_name_error=False,
            parser_type=None,
            pydantic_output_model=None,
            fallback_parser_type=None,
            fallback_pydantic_output_model=None
        )
    
    # Test invalid parser combination
    with pytest.raises(_ChainError):
        _ChainComposerValidator.validate_parser_configuration(
            output_passthrough_key_name="test",
            ignore_output_passthrough_key_name_error=False,
            parser_type=None,
            pydantic_output_model=None,
            fallback_parser_type="str",
            fallback_pydantic_output_model=None
        )
    
    # Test missing Pydantic model
    with pytest.raises(_ChainError):
        _ChainComposerValidator.validate_parser_configuration(
            output_passthrough_key_name="test",
            ignore_output_passthrough_key_name_error=False,
            parser_type="pydantic",
            pydantic_output_model=None,
            fallback_parser_type=None,
            fallback_pydantic_output_model=None
        )
    
    # Test invalid parser type
    with pytest.raises(_ChainError):
        _ChainComposerValidator.validate_parser_configuration(
            output_passthrough_key_name="test",
            ignore_output_passthrough_key_name_error=False,
            parser_type="invalid",
            pydantic_output_model=None,
            fallback_parser_type=None,
            fallback_pydantic_output_model=None
        )

def test_chain_composer_validator_prompt_variables():
    """Test ChainComposerValidator prompt variables validation"""
    # Test missing prompt variables
    with pytest.raises(_ChainError):
        _ChainComposerValidator.validate_prompt_variables(None, is_first_call=True)
    
    # Test invalid prompt variables type
    with pytest.raises(_ChainError):
        _ChainComposerValidator.validate_prompt_variables("not_a_dict", is_first_call=True)
    
    # Test valid prompt variables
    _ChainComposerValidator.validate_prompt_variables({"test": "value"}, is_first_call=True)
    _ChainComposerValidator.validate_prompt_variables(None, is_first_call=False) 


from enum import Enum

class _BaseErrorReference(Enum):
    pass

class _ChainBuilderErrorReference(_BaseErrorReference):
    pass

class _ChainManagerErrorReference(_BaseErrorReference):
    pass

class _ChainWrapperErrorReference(_BaseErrorReference):
    INPUT_ERROR = "input_error"
    MAIN_CHAIN_ERROR = "main_chain_error"
    FALLBACK_CHAIN_ERROR = "fallback_chain_error"

class _ChainComposerErrorReference(_BaseErrorReference):
    UNSUPPORTED_LLM_MODEL = "unsupported_llm_model"
    INVALID_PARSER_TYPE = "invalid_parser_type"
    MISSING_PYDANTIC_MODEL = "missing_pydantic_model"
    MISSING_OUTPUT_KEY = "missing_output_key"
    INVALID_PARSER_COMBINATION = "invalid_parser_combination"
    MISSING_PROMPT_VARIABLES = "missing_prompt_variables"
    INVALID_PROMPT_VARIABLES = "invalid_prompt_variables"
    INVALID_API_KEY = "invalid_api_key"
